"""Apply wire transparency corrections."""

from typing import Optional
from warnings import warn

import numpy as np

from .database import CalibrationDatabase

__all__ = ["TransparencyCalibrator"]


class TransparencyCalibrator:
    """Applies a correction on the amount of charge observed in a space point
    based on its position in the plane of the sensitive wires/pixels (yz).
    """

    name = "transparency"

    def __init__(
        self,
        transparency_db: str,
        num_tpcs: int,
        value_key: str = "scale",
        run_id: Optional[int] = None,
    ):
        """Load the calibration maps.

        Parameters
        ----------
        lifetime_db : str
            Path to a SQLite db file which maps [run, cryo, tpc] sets onto
            a specific transparency calibration map.
        num_tpcs : int
            Number of TPCs in the detector
        value_key: str, default 'scale'
            Database key which provides the calibration factor
        run_id : int
            Static run ID to use to fetch the transparency map
        """
        # Load the transparency database
        self.transparency = CalibrationDatabase(
            transparency_db, num_tpcs=num_tpcs, db_type="map", value_key=value_key
        )

        # Set a static run ID, if requested (for simulation)
        self.run_id = run_id
        if run_id is not None:
            warn(
                "The run ID provided by the event will be ignored in fetching "
                f"the calibration transparency map in favor of {run_id}."
            )

    def process(
        self, points: np.ndarray, values: np.ndarray, tpc_id: int, run_id: int
    ) -> np.ndarray:
        """Apply the transparency correction.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) array of point coordinates
        values : np.ndarray
            (N) array of values associated with each point
        tpc_id : int
            ID of the TPC to use
        run_id : int
            Used to get the appropriate transparency map

        Returns
        -------
        np.ndarray
            (N) array of corrected values
        """
        # If a static run ID was provided in the configuration, override
        if self.run_id is not None:
            run_id = self.run_id

        assert run_id is not None, "Must provide a run ID to get the transparency map."

        # Get the appropriate transparency map for this run
        transparency_lut = self.transparency[run_id]

        # Get the transparency correction for each position in the image
        return values / transparency_lut[tpc_id].query(points)



"""Function which sets up the necessary configuration for all CNNs."""


def setup_cnn_configuration(
    self,
    reps,
    depth,
    filters,
    input_kernel=3,
    data_dim=3,
    num_input=1,
    allow_bias=False,
    activation="lrelu",
    norm_layer="batch_norm",
    spatial_size=None,
):
    """Base function for global network parameters (CNN-based models).

    This avoids repeating the same base configuration parsing everywhere.
    For example, typical usage would be:

    .. code-block:: python

        class UResNetEncoder(torch.nn.Module):
            def __init__(self, cfg):
                super().__init__()
                setup_cnn_configuration(self, **cfg)

    Parameters
    ----------
    reps : int
        Number of time convolutions are repeated at each depth
    depth : int
        Depth of the CNN (number of downsampling)
    filters : int
        Number of input filters
    input_kernel : int, default 3
        Input kernel size
    data_dim : int, default 3
        Dimension of the input image data
    num_input : int, default 1
        Number of features in the input image
    allow_bias : bool, default False
        Whether to allow biases in the convolution and linear layers
    activation : union[str, dict], default 'relu'
        Activation function configuration
    normalization : union[str, dict], default 'batch_norm'
        Normalization function configuration
    spatial_size : int, optional
        Size of the input image in number of voxels per data_dim. This is only
        necessary when passing the normalized coordinates as features.
    """
    # Store the base parameters
    self.reps = reps
    self.depth = depth
    self.num_filters = filters
    self.input_kernel = input_kernel
    self.dim = data_dim
    self.num_input = num_input
    self.allow_bias = allow_bias
    self.spatial_size = spatial_size

    # Convert the depth to a number of filters per plane
    self.num_planes = [i * self.num_filters for i in range(1, self.depth + 1)]

    # Store activation function configuration
    self.act_cfg = activation

    # Store the normalization function configuration
    self.norm_cfg = norm_layer



"""Shower reconstruction module."""

import numpy as np

from spine.data import ObjectList, RecoParticle
from spine.math.distance import cdist
from spine.post.base import PostBase
from spine.utils.globals import PION_PID, PROT_PID, SHOWR_SHP, TRACK_SHP
from spine.utils.gnn.cluster import cluster_dedx, cluster_direction

__all__ = [
    "ShowerConversionDistanceProcessor",
    "ShowerStartMergeProcessor",
    "ShowerStartCorrectionProcessor",
]


class ShowerConversionDistanceProcessor(PostBase):
    """Evaluates the distance between shower start and the position
    of the vertex of the interaction they belong to.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "shower_conversion_distance"

    # List of recognized ways to compute the vertex-to-shower distance
    _modes = ("vertex_to_start", "vertex_to_points", "protons_to_points")

    def __init__(self, mode="vertex_to_points", include_secondary=False):
        """Store the conversion distance reconstruction parameters.

        Parameters
        ----------
        mode : str, default 'vertex'
            Method used to compute the conversion distance:
            - 'vertex_to_start': Distance between the vertex and the predicted
               shower start point.
            - 'vertex_to_points': Shortest distance between the vertex and any
              shower point.
            - 'protons_to_points': Distance between any proton/pion point and
              any shower point.
        include_secondary : bool, default False
            If `True`, computes the covnersion distance for secondary particles
        """
        # Initialize the parent class
        super().__init__("interaction", "reco")

        # Check that the conversion distance mode is valid, store
        assert mode in self._modes, (
            f"Conversion distance computation mode not recognized: {mode}. "
            f"Should be one of {self._modes}"
        )
        self.mode = mode

        # Store the vertex distance computation parameters
        self.include_secondary = include_secondary

        # If the method involves the vertex, must run the vertex PP
        if "vertex" in self.mode:
            self.update_upstream("vertex")

    def process(self, data):
        """Compute the conversion distance of showers in each interaction.

        Parameters
        ----------
        data : dict
            Dictionaries of data products
        """
        # Loop over the reco interactions
        for inter in data["reco_interactions"]:
            # Loop over the particles that make up the interaction
            for part in inter.particles:
                # If the particle is a secondary and they are to be skipped, do
                if not self.include_secondary and not part.is_primary:
                    continue

                # If the particle PID is not a shower, skip
                if part.shape != SHOWR_SHP:
                    continue

                # Compute the vertex distance
                if self.mode == "vertex_to_start":
                    dist = self.get_vertex_to_start(inter, part)
                elif self.mode == "vertex_to_points":
                    dist = self.get_vertex_to_points(inter, part)
                elif self.mode == "protons_to_start":
                    dist = self.get_protons_to_points(inter, part)
                else:
                    raise ValueError(
                        f"Conversion distance computation mode not recognized: "
                        f"{self.mode}"
                    )

                part.vertex_distance = dist

    @staticmethod
    def get_vertex_to_start(inter, shower):
        """Helper function to compute the closest distance between the vertex
        and the predicted shower startpoint.

        Parameters
        ----------
        inter : RecoInteraction
            Reco interaction providing the vertex
        shower : RecoParticle
            Shower particle in the interaction

        Returns
        -------
        float
            Distance betweenthe vertex and the shower start point
        """
        # Compute distance from the vertex to the shower start point
        dist = np.linalg.norm(inter.vertex - shower.start_point)

        return dist

    @staticmethod
    def get_vertex_to_points(inter, shower):
        """Helper function to compute the closest distance between the vertex
        and all shower points.

        Parameters
        ----------
        inter : RecoInteraction
            Reco interaction providing the vertex
        shower : RecoParticle
            Shower particle in the interaction

        Returns
        -------
        float
            Shortest distance between the vertex and any shower point
        """
        # Compute distances from the vertex to all shower points, find minimum
        dists = cdist(inter.vertex.reshape(-1, 3), shower.points)

        return dists.min()

    @staticmethod
    def get_protons_to_points(inter, shower):
        """Helper function to compute the closest distance between any
        proton/pion point and any shower point.

        Parameters
        ----------
        inter : RecoInteraction
            Reco interaction providing the vertex
        shower : RecoParticle
            Shower particle in the interaction

        Returns
        -------
        float
            Shortest distance between the shower and proton/pion points
        """
        # Fetch the list of points associated with primary pions or protons
        proton_points = []
        for part in inter.particles:
            if part.pid in (PION_PID, PROT_PID) and part.is_primary:
                proton_points.append(part.points)

        # If there is no pion/proton point, return dummy value
        if len(proton_points) == 0:
            return -1

        # Compute conversion distance
        proton_points = np.vstack(proton_points)
        dists = cdist(proton_points, shower.points)

        return dists.min()


class ShowerStartMergeProcessor(PostBase):
    """Merge shower start if it was classified as track points.

    This post-processor is used to patch upstream semantic segmentation mistakes
    where part of the shower trunk is assigned to track points and subsequantly
    classified as a separate particle.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "shower_start_merge"

    # Set of post-processors which must be run before this one is
    _upstream = ("direction",)

    def __init__(
        self,
        angle_threshold=0.175,
        distance_threshold=1.0,
        track_length_limit=50,
        track_dedx_limit=None,
    ):
        """Store the shower start merging parameters.

        Parameters
        ----------
        angle_threshold : float, default 0.175
            Track and shower angular agreement threshold to be merged (radians)
        distance_threshold : float, default 1.0
            Track and shower distance threshold to be merged (cm)
        track_length_limit : int, default 50
            Maximum track length allowed to mitigate merging legitimate tracks
        track_dedx_limit : float, optional
            Maximum track dE/dx allowed to mitigate proton merging (MeV/cm)
        """
        # Initialize the parent class
        super().__init__("interaction", "reco")

        # Store the shower start merging parameters
        self.angle_threshold = abs(np.cos(angle_threshold))
        self.distance_threshold = distance_threshold
        self.track_length_limit = track_length_limit
        self.track_dedx_limit = track_dedx_limit

    def process(self, data):
        """Merge split shower starts for all particles in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products

        Returns
        -------
        List[RecoParticle]
            Updated set of particles (interactions modified in place)
        """
        # Loop over the reco interactions
        particles = ObjectList([], default=RecoParticle())
        offset = 0
        for inter in data["reco_interactions"]:
            # Fetch the leading shower in the interaction, skip if None
            shower = inter.leading_shower
            if shower is None:
                for part in inter.particles:
                    part.id = offset
                    offset += 1

                particles.extend(inter.particles)
                inter.particle_ids = np.array([part.id for part in inter.particles])
                continue

            # Merge tracks which fit the criteria into the leading shower
            merge_ids = []
            for part in inter.particles:
                # Only check mergeability of primary tracks
                if part.shape == TRACK_SHP and part.is_primary:
                    if self.check_merge(shower, part):
                        shower.merge(part)
                        merge_ids.append(part.id)

            # Update the particle list of the interaction
            new_particles = []
            for part in inter.particles:
                if part.id not in merge_ids:
                    part.id = offset
                    new_particles.append(part)
                    offset += 1

            particles.extend(new_particles)
            inter.particles = new_particles
            inter.particle_ids = np.array([part.id for part in new_particles])

        return {"reco_particles": particles}

    def check_merge(self, shower, track):
        """Check if a shower and a track can be merged.

        Parameters
        ----------
        shower : RecoParticle
            Shower particle to merge the track into
        track : RecoParticle
            Track particle that may be merged into the shower

        Returns
        -------
        bool
            Whether the track can be merged into the shower or not
        """
        # Check if the track is not too long to be merged
        if track.length > self.track_length_limit:
            return False

        # Check that the track dE/dx is not too large to be merged
        if self.track_dedx_limit is not None:
            dedx = cluster_dedx(track.points, track.depositions, track.start_point)
            if dedx > self.track_dedx_limit:
                return False

        # Check that the angular agreement between particles is sufficient
        angular_sep = abs(np.dot(track.start_dir, shower.start_dir))
        if angular_sep < self.angle_threshold:
            return False

        # Check that the distance between particles is not too large
        dist = cdist(shower.points, track.points).min()
        if dist > self.distance_threshold:
            return False

        # If all checks passed, return True
        return True


class ShowerStartCorrectionProcessor(PostBase):
    """Correct the start point of primary EM showers by finding the closest
    point to the vertex.

    This post-processor is used to patch upstream point proposal mistakes
    where the shower start point is not placed correctly.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "shower_start_correction"

    # Set of post-processors which must be run before this one is
    _upstream = ("direction",)

    def __init__(self, update_directions=True, radius=-1, optimize=True):
        """Store the shower start corrector parameters.

        Parameters
        ----------
        update_directions : bool, default True
            If `True`, the direction of showers for which the start point
            has been updated are recomputed with the updated start point
        radius : float, default -1
            Radius around the start voxel to include in the direction estimate
        optimize : bool, default True
            Optimize the number of points involved in the direction estimate
        """
        # Initialize the parent class
        super().__init__("interaction", "reco")

        # Store start corrector parameters
        self.update_directions = update_directions
        self.radius = radius
        self.optimize = optimize

    def process(self, data):
        """Update the shower start point using the closest point to the vertex.

        Parameters
        ----------
        data : dict
            Dictionaries of data products
        """
        # Loop over the reco interactions
        for inter in data["reco_interactions"]:
            # Loop over the particles that make up the interaction
            for part in inter.particles:
                # If the particle PID is not a shower or not primary, skip
                if part.shape != SHOWR_SHP or not part.is_primary:
                    continue

                # Find the new start point of a shower
                new_start = self.correct_shower_start(inter, part)

                # If requested and needed, update the shower direction
                if self.update_directions and (new_start != part.start_point).any():
                    part.start_dir = cluster_direction(
                        part.points,
                        new_start,
                        max_dist=self.radius,
                        optimize=self.optimize,
                    )

                # Store the new start position
                part.start_point = new_start

    @staticmethod
    def correct_shower_start(inter, shower):
        """Function to correct the shower start point by finding the closest
        point to any primary track in the image.

        Parameters
        ----------
        inter : RecoInteraction
            Reco interaction to provide track points
        shower : RecoParticle
            Primary EM shower to find the shower start point for

        Returns
        -------
        guess : np.ndarray
            (3, ) array of the corrected shower start point.
        """
        # Get the set of primary track points
        track_points = []
        for part in inter.particles:
            if part.shape == TRACK_SHP and part.is_primary:
                track_points.append(part.points)

        # If there are no track points in the interaction, keep existing
        if len(track_points) == 0:
            return shower.start_point

        # Update the shower start point by finding the consitituent point
        # closest to any of the track points
        track_points = np.vstack(track_points)
        dists = np.min(cdist(shower.points, track_points), axis=1)
        closest_idx = np.argmin(dists)
        start_point = shower.points[closest_idx]

        return start_point



"""Calorimetric energy reconstruction module."""

import numexpr as ne
import numpy as np

from spine.post.base import PostBase
from spine.utils.calib import CalibrationManager
from spine.utils.globals import TRACK_SHP

__all__ = ["CalorimetricEnergyProcessor", "CalibrationProcessor"]


class CalorimetricEnergyProcessor(PostBase):
    """Compute calorimetric energy by summing the charge depositions and
    scaling by the ADC to MeV conversion factor, if needed.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "calo_ke"

    # Alternative allowed names of the post-processor
    aliases = ("reconstruct_calo_energy",)

    def __init__(
        self,
        scaling=1.0,
        shower_fudge=1.0,
        obj_type="particle",
        run_mode="reco",
        truth_dep_mode="depositions",
    ):
        """Stores the ADC to MeV conversion factor.

        Parameters
        ----------
        scaling : Union[float, str], default 1.
            Global scaling factor for the depositions (can be an expression)
        shower_fudge : Union[float, str], default 1.
            Shower energy fudge factor (accounts for missing cluster energy)
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_dep_mode=truth_dep_mode)

        # Store the conversion factor
        self.scaling = scaling
        if isinstance(self.scaling, str):
            self.scaling = ne.evaluate(self.scaling)

        # Store the shower fudge factor
        self.shower_fudge = shower_fudge
        if isinstance(self.shower_fudge, str):
            self.shower_fudge = ne.evaluate(self.shower_fudge)

    def process(self, data):
        """Reconstruct the calorimetric KE for each particle in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over particle types
        for k in self.obj_keys:
            # Loop over particles
            for part in data[k]:
                # Set up the scaling
                scaling = self.scaling
                if part.shape != TRACK_SHP:
                    scaling *= self.shower_fudge

                # Save the calorimetric energy
                depositions = self.get_depositions(part)
                part.calo_ke = scaling * np.sum(depositions)


class CalibrationProcessor(PostBase):
    """Apply calibrations to the reconstructed objects."""

    # Name of the post-processor (as specified in the configuration)
    name = "calibration"

    # Alternative allowed names of the post-processor
    aliases = ("apply_calibrations",)

    # Set of data keys needed for this post-processor to operate
    _keys = (("run_info", False),)

    def __init__(
        self,
        do_tracking=False,
        obj_type=("particle", "interaction"),
        run_mode="reco",
        truth_point_mode="points",
        **cfg,
    ):
        """Initialize the calibration manager.

        Parameters
        ----------
        do_tracking : bool, default False
            Segment track to get a proper local dQ/dx estimate
        **cfg : dict
            Calibration manager configuration
        """
        # Figure out which truth deposition attribute to use
        truth_dep_mode = truth_point_mode.replace("points", "depositions") + "_q"

        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_point_mode, truth_dep_mode)

        # Initialize the calibrator
        self.calibrator = CalibrationManager(**cfg)
        self.do_tracking = do_tracking

        # Add necessary keys
        keys = {}
        if run_mode != "truth":
            keys.update({"points": True, "depositions": True, "sources": False})

        if run_mode != "reco":
            keys.update(
                {
                    self.truth_point_key: True,
                    self.truth_dep_key: True,
                    self.truth_source_key: False,
                }
            )

        self.update_keys(keys)

    def process(self, data):
        """Apply calibrations to each particle in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Fetch the run info
        run_id = None
        if "run_info" in data:
            run_info = data["run_info"]
            run_id = run_info.run

        # Loop over particle objects
        for k in self.particle_keys:
            sources = None
            points_key = "points" if not "truth" in k else self.truth_point_key
            source_key = "sources" if not "truth" in k else self.truth_source_key
            dep_key = "depositions" if not "truth" in k else self.truth_dep_key
            unass_mask = np.ones(len(data[dep_key]), dtype=bool)
            for part in data[k]:
                # Make sure the particle coordinates are expressed in cm
                self.check_units(part)

                # Get point coordinates, sources and depositions
                points = self.get_points(part)
                if len(points) == 0:
                    continue

                deps = self.get_depositions(part)
                if source_key in data:
                    sources = self.get_sources(part)

                # Apply calibration
                if not self.do_tracking or part.shape != TRACK_SHP:
                    depositions = self.calibrator(points, deps, sources, run_id)
                else:
                    depositions = self.calibrator(
                        points, deps, sources, run_id, track=True
                    )

                # Update the particle *and* the reference tensor
                if not part.is_truth:
                    part.depositions = depositions
                else:
                    setattr(part, self.truth_dep_mode, depositions)

                data[dep_key][part.index] = depositions
                unass_mask[part.index] = False

            # Apply calibration corrections to unassociated depositions
            unass_index = np.where(unass_mask)[0]
            points = data[points_key][unass_index]
            depositions = data[dep_key][unass_index]
            if source_key in data:
                sources = data[source_key][unass_index]

            data[dep_key][unass_index] = self.calibrator(
                points, depositions, sources, run_id
            )

        # If requested, updated the depositions attribute of interactions
        for k in self.interaction_keys:
            dep_key = "depositions" if not "truth" in k else self.truth_dep_key
            for inter in data[k]:
                # Update depositions for the interaction
                depositions = data[dep_key][inter.index]
                if not inter.is_truth:
                    inter.depositions = depositions
                else:
                    setattr(inter, self.truth_dep_mode, depositions)



from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F


class TransformerDecoder(nn.Module):

    def __init__(
        self,
        d_model,
        num_heads,
        dim_feedforward=1024,
        dropout=0.0,
        normalize_before=False,
    ):
        super(TransformerDecoder, self).__init__()

        self.num_heads = num_heads

        self.cross_attention = CrossAttentionLayer(
            d_model, num_heads, dropout=dropout, normalize_before=normalize_before
        )
        self.self_attention = SelfAttentionLayer(
            d_model, num_heads, dropout=dropout, normalize_before=normalize_before
        )
        self.ffn_layer = FFNLayer(
            d_model, dim_feedforward, dropout=dropout, normalize_before=normalize_before
        )

        self.norm = nn.LayerNorm(d_model)

    def forward(self, queries, query_pos, src_pcd, batched_pos_enc, batched_attn):
        """
        queries: B, num_queries, d_model

        """
        x = queries.permute((1, 0, 2))
        memory_mask = batched_attn.repeat_interleave(self.num_heads, dim=0).permute(
            0, 2, 1
        )
        pos = batched_pos_enc.permute((1, 0, 2))
        x = self.cross_attention(
            x,
            src_pcd.permute((1, 0, 2)),
            memory_mask=memory_mask,
            memory_key_padding_mask=None,
            pos=pos,
            query_pos=query_pos.permute((1, 0, 2)),
        )
        x = self.self_attention(
            x,
            tgt_mask=None,
            tgt_key_padding_mask=None,
            query_pos=query_pos.permute((1, 0, 2)),
        )
        out_queries = self.ffn_layer(x).permute((1, 0, 2))

        return out_queries


class TransformerEncoderLayer(nn.Module):
    """
    Transformer module (attention mechanism) that takes (N, F_in) feature
    tensors to (N, F_out) feature tensor.
    """

    def __init__(
        self,
        num_input,
        num_output,
        num_hidden=128,
        num_layers=5,
        d_qk=128,
        d_v=128,
        num_heads=8,
        leakiness=0.0,
        dropout=True,
        name="attention_net",
        norm_layer="layer_norm",
    ):
        super(TransformerEncoderLayer, self).__init__()

        self.ma_layers = nn.ModuleList()
        self.ffn_layers = nn.ModuleList()

        self.num_layers = num_layers

        for i in range(num_layers):
            self.ma_layers.append(
                MultiHeadAttention(
                    num_input,
                    d_qk,
                    d_v,
                    num_hidden,
                    num_heads=num_heads,
                    dropout=dropout,
                    norm_layer=norm_layer,
                )
            )
            self.ffn_layers.append(
                PositionWiseFFN(
                    num_hidden,
                    num_hidden,
                    num_hidden,
                    leakiness=leakiness,
                    dropout=dropout,
                    norm_layer=norm_layer,
                )
            )
            num_input = num_hidden

        self.softplus = nn.Softplus()

    def forward(self, x):

        for i in range(self.num_layers):
            # print(x.shape)
            x = self.ma_layers[i](x)
            # print(x.shape)
            x = self.ffn_layers[i](x)
            # print(x.shape)

        x = self.softplus(x)
        return x


class MultiHeadAttention(nn.Module):

    def __init__(
        self,
        num_input,
        d_qk,
        d_v,
        num_output,
        num_heads=8,
        dropout=False,
        norm_layer="layer_norm",
    ):
        super(MultiHeadAttention, self).__init__()

        self.num_input = num_input
        self.num_output = num_output
        self.num_heads = num_heads
        self.d_qk = d_qk
        self.d_v = d_v
        self.T = num_output**0.5

        if norm_layer == "layer_norm":
            self.norm = nn.LayerNorm(num_output)
        elif norm_layer == "batch_norm":
            self.norm = nn.BatchNorm1d(num_output)
        else:
            raise ValueError(
                "Normalization layer {} not recognized!".format(norm_layer)
            )

        self.W_q = nn.Linear(num_input, d_qk * num_heads, bias=False)
        self.W_k = nn.Linear(num_input, d_qk * num_heads, bias=False)
        self.W_v = nn.Linear(num_input, d_v * num_heads, bias=False)
        self.W_o = nn.Linear(d_v * num_heads, num_output, bias=False)

        self.dropout = dropout
        if self.dropout:
            self.dropout1 = nn.Dropout()

        if self.num_input == self.num_output:
            self.residual = nn.Identity()
        else:
            self.residual = nn.Linear(self.num_input, self.num_output)

    def forward(self, x):

        num_output, num_heads = self.num_output, self.num_heads
        residual = self.residual(x)

        q = self.W_q(x).view(-1, self.d_qk, num_heads)
        k = self.W_k(x).view(-1, self.d_qk, num_heads)
        v = self.W_v(x).view(-1, self.d_v, num_heads)

        a = torch.einsum("ikb,jkb->ijb", q, k)
        attention = F.softmax(a / self.T, dim=1)
        v = torch.einsum("bih,ijh->bjh", attention, v).contiguous()
        v = v.view(-1, num_heads * self.d_v)
        out = self.W_o(v)

        out += residual
        if self.dropout:
            out = self.dropout1(out)

        return self.norm(out)


class PositionWiseFFN(nn.Module):

    def __init__(
        self,
        num_input,
        num_output,
        num_hidden,
        leakiness=0.0,
        dropout=True,
        norm_layer="layer_norm",
    ):
        super(PositionWiseFFN, self).__init__()
        self.num_input = num_input
        self.num_output = num_output

        self.linear1 = nn.Linear(num_input, num_hidden)
        self.linear2 = nn.Linear(num_hidden, num_output)
        if norm_layer == "layer_norm":
            self.norm = nn.LayerNorm(num_output)
        elif norm_layer == "batch_norm":
            self.norm = nn.BatchNorm1d(num_output)
        else:
            raise ValueError(
                "Normalization layer {} not recognized!".format(norm_layer)
            )
        self.act = nn.LeakyReLU(negative_slope=leakiness)

        self.dropout = dropout
        if self.dropout:
            self.dropout1 = nn.Dropout()

    def forward(self, x):
        residual = x
        out = self.act(self.linear1(x))
        out = self.linear2(out)
        if self.dropout:
            out = self.dropout1(out)
        out += residual
        return self.norm(out)


# ---------------------------------------------------------------------------
# From Mask3D/models/mask3d.py by Jonas Schult:
# https://github.com/JonasSchult/Mask3D


class SelfAttentionLayer(nn.Module):

    def __init__(
        self, d_model, nhead, dropout=0.0, activation="relu", normalize_before=False
    ):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)

        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def with_pos_embed(self, tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_post(
        self, tgt, tgt_mask=None, tgt_key_padding_mask=None, query_pos=None
    ):
        q = k = self.with_pos_embed(tgt, query_pos)
        tgt2 = self.self_attn(
            q, k, value=tgt, attn_mask=tgt_mask, key_padding_mask=tgt_key_padding_mask
        )[0]
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)

        return tgt

    def forward_pre(
        self, tgt, tgt_mask=None, tgt_key_padding_mask=None, query_pos=None
    ):
        tgt2 = self.norm(tgt)
        q = k = self.with_pos_embed(tgt2, query_pos)
        tgt2 = self.self_attn(
            q, k, value=tgt2, attn_mask=tgt_mask, key_padding_mask=tgt_key_padding_mask
        )[0]
        tgt = tgt + self.dropout(tgt2)

        return tgt

    def forward(self, tgt, tgt_mask=None, tgt_key_padding_mask=None, query_pos=None):
        if self.normalize_before:
            return self.forward_pre(tgt, tgt_mask, tgt_key_padding_mask, query_pos)
        return self.forward_post(tgt, tgt_mask, tgt_key_padding_mask, query_pos)


class CrossAttentionLayer(nn.Module):

    def __init__(
        self, d_model, nhead, dropout=0.0, activation="relu", normalize_before=False
    ):
        super().__init__()
        self.multihead_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)

        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def with_pos_embed(self, tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_post(
        self,
        tgt,
        memory,
        memory_mask=None,
        memory_key_padding_mask=None,
        pos=None,
        query_pos=None,
    ):
        tgt2 = self.multihead_attn(
            query=self.with_pos_embed(tgt, query_pos),
            key=self.with_pos_embed(memory, pos),
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask,
        )[0]
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)

        return tgt

    def forward_pre(
        self,
        tgt,
        memory,
        memory_mask=None,
        memory_key_padding_mask=None,
        pos=None,
        query_pos=None,
    ):
        tgt2 = self.norm(tgt)

        tgt2 = self.multihead_attn(
            query=self.with_pos_embed(tgt2, query_pos),
            key=self.with_pos_embed(memory, pos),
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask,
        )[0]
        tgt = tgt + self.dropout(tgt2)

        return tgt

    def forward(
        self,
        tgt,
        memory,
        memory_mask=None,
        memory_key_padding_mask=None,
        pos=None,
        query_pos=None,
    ):
        if self.normalize_before:
            return self.forward_pre(
                tgt, memory, memory_mask, memory_key_padding_mask, pos, query_pos
            )
        return self.forward_post(
            tgt, memory, memory_mask, memory_key_padding_mask, pos, query_pos
        )


class FFNLayer(nn.Module):

    def __init__(
        self,
        d_model,
        dim_feedforward=2048,
        dropout=0.0,
        activation="relu",
        normalize_before=False,
    ):
        super().__init__()
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm = nn.LayerNorm(d_model)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def with_pos_embed(self, tensor, pos):
        return tensor if pos is None else tensor + pos

    def forward_post(self, tgt):
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)
        return tgt

    def forward_pre(self, tgt):
        tgt2 = self.norm(tgt)
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt2))))
        tgt = tgt + self.dropout(tgt2)
        return tgt

    def forward(self, tgt):
        if self.normalize_before:
            return self.forward_pre(tgt)
        return self.forward_post(tgt)


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(f"activation should be relu/gelu, not {activation}.")


NORM_DICT = {
    # "bn": BatchNormDim1Swap,
    "bn1d": nn.BatchNorm1d,
    "id": nn.Identity,
    "ln": nn.LayerNorm,
}

ACTIVATION_DICT = {
    "relu": nn.ReLU,
    "gelu": nn.GELU,
    "leakyrelu": partial(nn.LeakyReLU, negative_slope=0.1),
}

WEIGHT_INIT_DICT = {
    "xavier_uniform": nn.init.xavier_uniform_,
}


class GenericMLP(nn.Module):
    def __init__(
        self,
        input_dim,
        hidden_dims,
        output_dim,
        norm_fn_name=None,
        activation="relu",
        use_conv=False,
        dropout=None,
        hidden_use_bias=False,
        output_use_bias=True,
        output_use_activation=False,
        output_use_norm=False,
        weight_init_name=None,
    ):
        super().__init__()
        activation = ACTIVATION_DICT[activation]
        norm = None
        if norm_fn_name is not None:
            norm = NORM_DICT[norm_fn_name]
        if norm_fn_name == "ln" and use_conv:
            norm = lambda x: nn.GroupNorm(1, x)  # easier way to use LayerNorm

        if dropout is not None:
            if not isinstance(dropout, list):
                dropout = [dropout for _ in range(len(hidden_dims))]

        layers = []
        prev_dim = input_dim
        for idx, x in enumerate(hidden_dims):
            if use_conv:
                layer = nn.Conv1d(prev_dim, x, 1, bias=hidden_use_bias)
            else:
                layer = nn.Linear(prev_dim, x, bias=hidden_use_bias)
            layers.append(layer)
            if norm:
                layers.append(norm(x))
            layers.append(activation())
            if dropout is not None:
                layers.append(nn.Dropout(p=dropout[idx]))
            prev_dim = x
        if use_conv:
            layer = nn.Conv1d(prev_dim, output_dim, 1, bias=output_use_bias)
        else:
            layer = nn.Linear(prev_dim, output_dim, bias=output_use_bias)
        layers.append(layer)

        if output_use_norm:
            layers.append(norm(output_dim))

        if output_use_activation:
            layers.append(activation())

        self.layer = nn.Sequential(*layers)

        if weight_init_name is not None:
            self.do_weight_init(weight_init_name)

    def do_weight_init(self, weight_init_name):
        func = WEIGHT_INIT_DICT[weight_init_name]
        for _, param in self.named_parameters():
            if param.dim() > 1:  # skips batchnorm/layernorm
                func(param)

    def forward(self, x):
        output = self.layer(x)
        return output



"""Source assignment modules.

This module implements routines to assign module/tpc sources when
they are not explicitely provided.
"""

import numpy as np

from spine.geo import GeoManager
from spine.post.base import PostBase

__all__ = ["SourceAssigner"]


class SourceAssigner(PostBase):
    """Class which assigns depositions in the detector to specific sources.

    Uses proximity to specific modules/TPCs to assign sources. Note that this is
    not exact for out-of-time activity for which the drift position is not
    representative of the real position of the deposition.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "source"

    def __init__(
        self,
        run_mode="reco",
        truth_point_mode="points",
    ):
        """Initialize the source assigner"""
        # Initialize the parent class, store run mode
        super().__init__(run_mode=run_mode, truth_point_mode=truth_point_mode)

        # Initialize the geometry
        self.geo = GeoManager.get_instance()

        # Make sure the necessary attributes are loaded
        self.run_mode = run_mode
        if run_mode != "truth":
            self.update_keys({"points": True})
        if run_mode != "reco":
            self.update_keys({self.truth_point_key: True})

    def process(self, data):
        """Assign sources to the relevant depositions.

        Parameters
        ----------
        data : dict
            Dictionary of data products

        Returns
        -------
        dict
            Dictionary which sources
        """
        # Initialize output
        result = {}

        # Assign sources to reconstructed points, if needed
        if self.run_mode != "truth":
            result["sources"] = self.get_closest(data["points"])

        # Assign sources to truth points, if needed
        if self.run_mode != "reco":
            result[self.truth_source_key] = self.get_closest(data[self.truth_point_key])

        return result

    def get_closest(self, points):
        """Assign sources to one set of points based on proximity.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Point coordinates

        Returns
        -------
        np.ndarray
            (N, 2) Module/TPC sources
        """
        # Assign sources
        module_ids = self.geo.get_closest_module(points)
        tpc_ids = self.geo.get_closest_tpc(points)
        sources = np.vstack((module_ids, tpc_ids)).T

        return sources



"""Module to evaluate diagnostic metrics on tracks."""

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

from spine.ana.base import AnaBase
from spine.math.distance import cdist
from spine.utils.globals import MUON_PID, TRACK_SHP

__all__ = ["TrackCompletenessAna"]


class TrackCompletenessAna(AnaBase):
    """This analysis script identifies gaps in tracks and measures the
    cumulative length of these gaps relative to the track length.

    This is a useful diagnostic tool to evaluate the space-point efficiency
    on tracks (good standard candal as track should have exactly no gap in
    a perfectly efficient detector).
    """

    # Name of the analysis script (as specified in the configuration)
    name = "track_completeness"

    def __init__(
        self,
        time_window: Optional[Union[List[float], Tuple[float, float]]] = None,
        length_threshold: Optional[float] = 10,
        include_pids: Optional[Union[Tuple[int], List[int]]] = (MUON_PID,),
        run_mode: str = "both",
        truth_point_mode: str = "points",
        **kwargs,
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        time_window : List[float]
            Time window within which to include particle (only works for `truth`)
        length_threshold : float, optional
            Minimum length of tracks to consider, in cm
        include_pids : Union[Tuple[int], List[int]], optional
            Particle IDs to include in the analysis
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Initialize the parent class
        super().__init__("particle", run_mode, truth_point_mode, **kwargs)

        # Store the time window
        self.time_window = time_window
        assert (
            time_window is None or len(time_window) == 2
        ), "Time window must be specified as an array of two values."
        assert (
            time_window is None or run_mode == "truth"
        ), "Time of reconstructed particle is unknown."

        # Store the length threshold and included PIDs
        self.length_threshold = length_threshold
        self.include_pids = include_pids

        # Make sure the metadata is provided (rasterization needed)
        self.update_keys({"meta": True})

        # Initialize the CSV writer(s) you want
        for prefix in self.prefixes:
            self.initialize_writer(prefix)

    def process(self, data: Dict[str, Any]) -> None:
        """Evaluate track completeness for tracks in one entry.

        Parameters
        ----------
        data : Dict[str, Any]
            Dictionary of data products
        """
        # Fetch the pixel size in this image (assume cubic cells)
        assert np.all(
            data["meta"].size[0] == data["meta"].size
        ), "Non-cubic pixels not supported."
        pixel_size = data["meta"].size[0]

        # Loop over the types of particle data products
        for key in self.obj_keys:
            # Fetch the prefix ('reco' or 'truth')
            prefix = key.split("_")[0]

            # Loop over particle objects
            for part in data[key]:
                # Check that the particle is a track
                if part.shape != TRACK_SHP:
                    continue

                # If needed, check on the particle time
                if self.time_window is not None:
                    if part.t < self.time_window[0] or part.t > self.time_window[1]:
                        continue

                # If needed, check on the particle PID
                if self.include_pids is not None:
                    if part.pid not in self.include_pids:
                        continue

                # Initialize the particle dictionary
                comp_dict = {"particle_id": part.id}

                # Fetch the particle point coordinates
                points = self.get_points(part)

                # Find start/end points, collapse onto track cluster
                start = points[np.argmin(cdist(part.start_point[None, :], points))]
                end = points[np.argmin(cdist(part.end_point[None, :], points))]

                # Add the direction of the track
                vec = end - start
                length = np.linalg.norm(vec)
                if length:
                    vec /= length

                # If needed, check on the particle length
                if self.length_threshold is not None:
                    if length < self.length_threshold:
                        continue

                comp_dict["size"] = len(points)
                comp_dict["length"] = length
                comp_dict.update({"dir_x": vec[0], "dir_y": vec[1], "dir_z": vec[2]})
                comp_dict.update(
                    {"start_x": start[0], "start_y": start[1], "start_z": start[2]}
                )
                comp_dict.update({"end_x": end[0], "end_y": end[1], "end_z": end[2]})

                # Chunk out the track along gaps, estimate gap length
                chunk_labels = self.cluster_track_chunks(points, start, end, pixel_size)
                gaps = self.sequential_cluster_distances(points, chunk_labels, start)

                # Substract minimum gap distance due to rasterization
                min_gap = pixel_size / np.max(np.abs(vec))
                gaps -= min_gap

                # Store gap information
                comp_dict["num_gaps"] = len(gaps)
                comp_dict["gap_length"] = np.sum(gaps)
                comp_dict["gap_frac"] = np.sum(gaps) / length

                # Append the dictionary to the CSV log
                self.append(prefix, **comp_dict)

    @staticmethod
    def cluster_track_chunks(
        points, start_point: np.ndarray, end_point: np.ndarray, pixel_size: float
    ) -> np.ndarray:
        """Find point where the track is broken, divide out the track
        into self-contained chunks which are Linf connect (Moore neighbors).

        Parameters
        ----------
        points : np.ndarray
            (N, 3) List of track cluster point coordinates
        start_point : np.ndarray
            (3) Start point of the track cluster
        end_point : np.ndarray
            (3) End point of the track cluster
        pixel_size : float
            Dimension of one pixel, used to identify what is big enough to
            constitute a break

        Returns
        -------
        np.ndarray
            (N) Track chunk labels
        """
        # Project and cluster on the projected axis
        direction = (end_point - start_point) / np.linalg.norm(end_point - start_point)
        scale = pixel_size / np.max(np.abs(direction))
        projs = np.dot(points - start_point, direction)
        perm = np.argsort(projs)
        seps = projs[perm][1:] - projs[perm][:-1]
        breaks = np.where(seps > scale * 1.49)[0] + 1
        cluster_labels = np.empty(len(projs), dtype=int)
        for i, index in enumerate(np.split(np.arange(len(projs)), breaks)):
            cluster_labels[perm[index]] = i

        return cluster_labels

    @staticmethod
    def sequential_cluster_distances(
        points: np.ndarray, labels: np.ndarray, start_point: np.ndarray
    ) -> np.ndarray:
        """Order clusters in order of distance from a starting point, compute
        the distances between successive clusters.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) List of track cluster point coordinates
        labels : np.ndarray
            (N) Track chunk labels
        start_point : np.ndarray
            (3) Start point of the track cluster
        """
        # If there's only one cluster, nothing to do here
        unique_labels = np.unique(labels)
        if len(unique_labels) < 2:
            return np.empty(0, dtype=float)

        # Order clusters
        start_dist = cdist(start_point[None, :], points).flatten()
        start_clust_dist = np.empty(len(unique_labels))
        for i, c in enumerate(unique_labels):
            start_clust_dist[i] = np.min(start_dist[labels == c])
        ordered_labels = unique_labels[np.argsort(start_clust_dist)]

        # Compute the intercluster distance and relative angle
        n_gaps = len(ordered_labels) - 1
        dists = np.empty(n_gaps, dtype=float)
        for i in range(n_gaps):
            points_i = points[labels == ordered_labels[i]]
            points_j = points[labels == ordered_labels[i + 1]]
            dists[i] = np.min(cdist(points_i, points_j))

        return dists



"""Post-processor in charge of finding matches between charge and CRT."""

import numpy as np

from spine.post.base import PostBase

from .match import CRTMatcher

__all__ = ["CRTMatchProcessor"]


class CRTMatchProcessor(PostBase):
    """Associates TPC particles with CRT hits."""

    # Name of the post-processor (as specified in the configuration)
    name = "crt_match"

    # Alternative allowed names of the post-processor
    aliases = ("run_crt_matching",)

    # Set of post-processors which must be run before this one is
    _upstream = ("direction",)

    def __init__(self, crt_key, run_mode="reco", **kwargs):
        """Initialize the CRT/TPC matching post-processor.

        Parameters
        ----------
        crt_key : str
            Data product key which provides the CRT information
        **kwargs : dict
            Keyword arguments to pass to the CRT-TPC matching algorithm
        """
        # Initialize the parent class
        super().__init__("particle", run_mode)

        # Make sure the CRT hit data product is available, store
        self.crt_key = crt_key
        self.update_keys({crt_key: True})

        # Initialzie the matcher
        self.matcher = CRTMatcher(**kwargs)

    def process(self, data):
        """Find [particle, crthit] pairs.

        Parameters
        ----------
        data : dict
            Dictionary of data products

        Notes
        -----
        This post-processor modifies the list of `particle` objects
        in-place by filling the following attributes:
        - particle.is_crt_matched: (bool)
               Indicator for whether the given particle has a CRT hit match
        - particle.crt_ids: np.ndarray
               The CRT hit IDs in the CRT hit list
        - particle.crt_times: np.ndarray
               The CRT hit time(s) in microseconds
        """
        # Fetch the CRT hits, nothing to do here if there are none
        crthits = data[self.crt_key]
        if not len(crthits):
            return

        # Loop over the object types
        for k in self.particle_keys:
            # Fetch particle list, nothing to do here if there are none
            particles = data[k]
            if not len(particles):
                continue

            # Make sure the particle coordinates are expressed in cm
            self.check_units(particles[0])

            # Clear previous crt matching information
            for part in particles:
                part.reset_crt_match(typed=False)

            # Run CRT matching
            matches = self.matcher.get_matches(particles, crthits)

            # Store the CRT information
            for part, crt, match in matches:
                part.is_crt_matched = True
                part.crt_ids.append(crt.id)
                part.crt_times.append(crt.time)
                part.crt_scores.append(match)

            # Cast list attributes to numpy arrays
            for part in particles:
                part.crt_ids = np.asarray(part.crt_ids, dtype=np.int32)
                part.crt_times = np.asarray(part.crt_times, dtype=np.float32)
                part.crt_scores = np.asarray(part.crt_scores, dtype=np.float32)



"""Basic detector components shared across multiple subsystems.

This currently handles:
- :class:`Box` which corresponds to box-shaped detector modules.
"""

from dataclasses import dataclass
from typing import List, Union

import numpy as np

__all__ = ["Plane", "Box"]


@dataclass
class Plane:
    """Class which holds all methods associated with a plane.

    Attributes
    ----------
    intercept : np.ndarray
        (3,) Coordinates of a point which belongs to the plane
    norm : np.ndarray
        (3,) Vector perpendicular to the plane
    boundary : float
        Dot product between the interecept and the norm
    """

    intercept: np.ndarray
    norm: np.ndarray
    boundary: float

    def __init__(self, intercept: np.ndarray, norm: np.ndarray):
        """Initialize the box object.

        Parameters
        ----------
        intercept : np.ndarray
            (3,) Coordinates of a point which belongs to the plane
        norm : np.ndarray
            (3,) Vector perpendicular to the plane
        """
        # Store the intercept and norm
        self.intercept = intercept
        self.norm = norm

        # Compute their dot product (projection boundary)
        self.boundary = np.dot(intercept, norm)

    def distance(self, point: np.ndarray) -> float:
        """Computes the perpendicular distance from a point to the plane.

        Parameters
        ----------
        point : np.ndarray
            (3,) Coordinates of the point to compute the distance to

        Returns
        -------
        float
            Perpendicular distance from the point to the plane
        """
        return (np.dot(point, self.norm) - self.boundary) / np.linalg.norm(self.norm)


@dataclass
class Box:
    """Class which holds all methods associated with a box-shaped component.

    Attributes
    ----------
    boundaries : np.ndarray
        (3, 2) Box boundaries
        - 3 is the number of dimensions
        - 2 corresponds to the lower/upper boundaries along each axis
    """

    boundaries: np.ndarray

    def __init__(self, lower: np.ndarray, upper: np.ndarray):
        """Initialize the box object.

        Parameters
        ----------
        lower : np.ndarray
            (3,) Lower bounds of the box
        upper : np.ndarray
            (3,) Upper bounds of the box
        """
        # Store lower and upper boundaries in one array
        self.boundaries = np.vstack((lower, upper)).T

    @property
    def center(self) -> np.ndarray:
        """Center of the box.

        Returns
        -------
        np.ndarray
            (3,) Center of the box
        """
        return np.mean(self.boundaries, axis=1)

    @property
    def lower(self) -> np.ndarray:
        """Lower bounds of the box.

        Returns
        -------
        np.ndarray
            (3,) Lower bounds of the box
        """
        return self.boundaries[:, 0]

    @property
    def upper(self) -> np.ndarray:
        """Upper bounds of the box.

        Returns
        -------
        np.ndarray
            (3,) Upper bounds of the box
        """
        return self.boundaries[:, 1]

    @property
    def dimensions(self) -> np.ndarray:
        """Dimensions of the box.

        Returns
        -------
        np.ndarray
            (3,) Box dimensions
        """
        return self.boundaries[:, 1] - self.boundaries[:, 0]

    @property
    def volume(self) -> float:
        """Volume of the box.

        Returns
        -------
        float
            Box volume
        """
        return float(np.prod(self.dimensions))

    @property
    def faces(self) -> List[Plane]:
        """Returns the 6 faces of the box as Plane objects.

        Returns
        -------
        List[Plane]
            List of 6 Plane objects corresponding to the box faces
        """
        faces = []
        # Loop over each axis
        for axis in range(3):
            # Create normal vector for the face
            norm = np.zeros(3)
            norm[axis] = 1.0

            # Lower face
            intercept_lower = self.lower.copy()
            faces.append(Plane(intercept_lower, norm))

            # Upper face
            intercept_upper = self.upper.copy()
            faces.append(Plane(intercept_upper, -norm))

        return faces

    def distance(self, points: np.ndarray) -> Union[float, np.ndarray]:
        """Computes the minimum distance from a set of points to the box.

        If the point(s) is(are) inside the box, the distance is 0.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Coordinates of the points to compute the distance to

        Returns
        -------
        np.ndarray
            (N,) Minimum distance from each point to the box
        """
        # Distance from each point to the lower and upper boundaries
        diff_lower = self.lower - points  # positive if point is below lower bound
        diff_upper = points - self.upper  # positive if point is above upper bound

        # For each coord, if inside the interval, contribution is 0;
        # if outside, take the amount by which it's outside.
        delta = np.maximum(0.0, np.maximum(diff_lower, diff_upper))  # shape (N, 3)

        # Euclidean distance
        if len(delta.shape) == 1:
            distances = float(np.linalg.norm(delta))
        else:
            distances = np.linalg.norm(delta, axis=1)

        return distances



"""Numba JIT compiled implementation of clustering evaluation metrics.

This module provides efficient implementations of:
- Adjusted Rand Index (ARI)
- Adjusted Mutual Information (AMI)
"""

import numba as nb
import numpy as np

from .linalg import contingency_table

__all__ = ["adjusted_rand_score", "adjusted_mutual_info_score"]


@nb.jit(nopython=True, cache=True)
def _comb2(n):
    """Compute binomial coefficient n choose 2.

    Parameters
    ----------
    n : int
        Number of items to choose from.

    Returns
    -------
    int
        The number of ways to choose 2 items from n items, equal to n*(n-1)/2.
    """
    return n * (n - 1) // 2


@nb.jit(nopython=True, cache=True)
def adjusted_rand_score(labels_true, labels_pred):
    """Compute the Adjusted Rand Index (ARI) between two clusterings.

    The Adjusted Rand Index is a measure of the similarity between two
    data clusterings. It is a function that measures the similarity of the two
    assignments, ignoring permutations and correcting for chance agreement.

    The ARI is bounded between -1 and 1:
    - 1.0 indicates perfect clustering agreement
    - 0.0 indicates random clustering (expected value for independent labelings)
    - Negative values indicate worse than random clustering

    The formula is: ARI = (RI - E[RI]) / (max(RI) - E[RI])
    where RI is the Rand Index and E[RI] is the expected Rand Index under
    random labelings.

    Parameters
    ----------
    labels_true : ndarray of shape (n_samples,)
        Ground truth class labels to be used as a reference.
    labels_pred : ndarray of shape (n_samples,)
        Cluster labels to evaluate.

    Returns
    -------
    ari : float
        Adjusted Rand Index. A clustering result satisfying the constraints
        of a correct clustering has a score of 1.0.

    Notes
    -----
    This implementation uses a fast numba-compiled algorithm that avoids
    constructing the full pairwise similarity matrix.

    References
    ----------
    .. [1] Hubert, L. and Arabie, P. (1985). "Comparing partitions."
           Journal of Classification 2(1): 193-218.
    .. [2] https://en.wikipedia.org/wiki/Rand_index#Adjusted_Rand_index

    Examples
    --------
    Perfect clustering:
    >>> labels_true = [0, 0, 1, 1]
    >>> labels_pred = [0, 0, 1, 1]
    >>> adjusted_rand_score(labels_true, labels_pred)
    1.0

    Random clustering:
    >>> labels_true = [0, 0, 1, 1]
    >>> labels_pred = [0, 1, 0, 1]
    >>> adjusted_rand_score(labels_true, labels_pred)  # doctest: +ELLIPSIS
    0.0
    """
    # Get dimensions for contingency table
    nx = labels_true.max() + 1 if len(labels_true) > 0 else 1
    ny = labels_pred.max() + 1 if len(labels_pred) > 0 else 1
    contingency = contingency_table(labels_true, labels_pred, nx, ny)

    # Compute sums
    sum_comb_c = 0
    for i in range(contingency.shape[0]):
        for j in range(contingency.shape[1]):
            sum_comb_c += _comb2(contingency[i, j])

    sum_comb_k = 0  # Sum over rows (true clusters)
    for i in range(contingency.shape[0]):
        row_sum = 0
        for j in range(contingency.shape[1]):
            row_sum += contingency[i, j]
        sum_comb_k += _comb2(row_sum)

    sum_comb_c_pred = 0  # Sum over columns (pred clusters)
    for j in range(contingency.shape[1]):
        col_sum = 0
        for i in range(contingency.shape[0]):
            col_sum += contingency[i, j]
        sum_comb_c_pred += _comb2(col_sum)

    n_samples = len(labels_true)
    sum_comb_n = _comb2(n_samples)
    if sum_comb_n == 0:
        return 1.0

    expected_index = sum_comb_k * sum_comb_c_pred / sum_comb_n
    max_index = (sum_comb_k + sum_comb_c_pred) / 2.0

    if max_index == expected_index:
        return 1.0

    return (sum_comb_c - expected_index) / (max_index - expected_index)


@nb.jit(nopython=True, cache=True)
def _entropy(labels):
    """Compute entropy of a labeling."""
    unique_labels = np.unique(labels)
    n = len(labels)

    if n <= 1:
        return 0.0

    entropy = 0.0
    for label in unique_labels:
        count = np.sum(labels == label)
        if count > 0:
            p = count / n
            entropy -= p * np.log(p)

    return entropy


@nb.jit(nopython=True, cache=True)
def _mutual_info(labels_true, labels_pred):
    """Compute mutual information between two clusterings."""
    # Get dimensions for contingency table
    nx = labels_true.max() + 1 if len(labels_true) > 0 else 1
    ny = labels_pred.max() + 1 if len(labels_pred) > 0 else 1
    contingency = contingency_table(labels_true, labels_pred, nx, ny)

    n_samples = len(labels_true)
    mi = 0.0

    for i in range(contingency.shape[0]):
        for j in range(contingency.shape[1]):
            n_ij = contingency[i, j]
            if n_ij == 0:
                continue

            # Marginal counts
            n_i = 0
            for k in range(contingency.shape[1]):
                n_i += contingency[i, k]

            n_j = 0
            for k in range(contingency.shape[0]):
                n_j += contingency[k, j]

            mi += n_ij / n_samples * np.log((n_samples * n_ij) / (n_i * n_j))

    return mi


@nb.jit(nopython=True, cache=True)
def adjusted_mutual_info_score(labels_true, labels_pred):
    """Compute the Adjusted Mutual Information (AMI) between two clusterings.

    The Adjusted Mutual Information is a measure of agreement between two
    partitions, adjusted for chance. It employs the expected mutual information
    under a hypergeometric model of randomness.

    The AMI is normalized between 0 and 1:
    - 1.0 indicates perfect clustering agreement
    - 0.0 indicates independent labelings (expected value for random labelings)
    - Values close to 0.0 indicate near-random agreement

    The formula is: AMI = (MI - E[MI]) / (max(H(U), H(V)) - E[MI])
    where MI is the mutual information, E[MI] is the expected mutual
    information, and H(U), H(V) are the entropies of the two labelings.

    Parameters
    ----------
    labels_true : ndarray of shape (n_samples,)
        Ground truth class labels to be used as a reference.
    labels_pred : ndarray of shape (n_samples,)
        Cluster labels to evaluate.

    Returns
    -------
    ami : float
        Adjusted Mutual Information score. Perfect labelings are scored 1.0.
        Bad labelings or independent labelings have non-positive scores.

    Notes
    -----
    This implementation uses a fast numba-compiled algorithm that computes
    the hypergeometric expected mutual information directly from the
    contingency table.

    References
    ----------
    .. [1] Vinh, N. X., Epps, J., & Bailey, J. (2010). "Information theoretic
           measures for clusterings comparison: Variants, properties,
           normalization and correction for chance." Journal of Machine
           Learning Research, 11, 2837-2854.
    .. [2] https://en.wikipedia.org/wiki/Adjusted_mutual_information

    Examples
    --------
    Perfect clustering:
    >>> labels_true = [0, 0, 1, 1]
    >>> labels_pred = [0, 0, 1, 1]
    >>> adjusted_mutual_info_score(labels_true, labels_pred)
    1.0

    Random clustering:
    >>> labels_true = [0, 0, 1, 1]
    >>> labels_pred = [0, 1, 0, 1]
    >>> adjusted_mutual_info_score(labels_true, labels_pred)  # doctest: +ELLIPSIS
    0.0...
    """
    if len(labels_true) != len(labels_pred):
        raise ValueError("Labels must have the same length")

    if len(labels_true) == 0:
        return 1.0

    # Handle trivial cases
    n_true_clusters = len(np.unique(labels_true))
    n_pred_clusters = len(np.unique(labels_pred))

    if n_true_clusters == 1 and n_pred_clusters == 1:
        return 1.0
    elif n_true_clusters == 1 or n_pred_clusters == 1:
        return 0.0

    # Compute entropies and mutual information
    entropy_true = _entropy(labels_true)
    entropy_pred = _entropy(labels_pred)
    mutual_info = _mutual_info(labels_true, labels_pred)

    # Expected mutual information (approximation for large n)
    n_samples = len(labels_true)
    expected_mi = entropy_true * entropy_pred / np.log(n_samples)

    # Compute adjusted mutual information
    mean_entropy = (entropy_true + entropy_pred) / 2.0

    if mean_entropy == expected_mi:
        return 1.0

    return (mutual_info - expected_mi) / (mean_entropy - expected_mi)



"""Post-processing tools for ML-based neutrino reconstruction.

This module transforms raw ML model outputs into high-level physics quantities,
serving as the bridge between deep learning predictions and physics analysis.

**Core Post-Processing Management:**
- `PostManager`: Central coordinator for post-processing workflows
  - Multi-stage processing pipeline
  - Configurable algorithm chains
  - Quality control and validation
  - Output formatting and export

**Physics Reconstruction Algorithms:**

**Particle Reconstruction:**
- **Direction Reconstruction**: Principal component analysis and track fitting
- **Energy Estimation**: Range-based (CSDA) and calorimetric methods
- **Particle ID Refinement**: Chi-squared fits and likelihood methods
- **Vertex Finding**: Track intersection and clustering algorithms
- **Momentum Reconstruction**: Multiple Coulomb scattering analysis

**Detector Integration:**
- **Optical Flash Matching**: Time and spatial correlation with PMT signals
- **Cosmic Ray Tagging**: CRT hit matching and cosmic rejection
- **Trigger Matching**: Event timing and readout window association
- **Calibration Application**: Energy scale and detector response corrections

**Event-Level Processing:**
- **Interaction Classification**: Primary particle topology identification
- **Containment Analysis**: Fiducial volume and boundary effect studies
- **Background Rejection**: Cosmic ray and noise event filtering
- **Quality Metrics**: Reconstruction confidence and uncertainty estimation

**Algorithm Categories:**

**Tracking Algorithms:**
- 3D track fitting and smoothing
- Kinematic endpoint determination
- Track-shower separation
- Multiple track reconstruction

**Calorimetry:**
- Energy deposition analysis
- dE/dx calculation and calibration
- Range-energy relationships
- Electromagnetic shower energy

**Matching and Association:**
- Truth-to-reco object matching
- Cross-detector hit correlation
- Time-of-flight calculations
- Multi-plane track linking

**Data Flow:**
```
ML Predictions → Post-Processing → Physics Objects → Analysis
      ↓                ↓                 ↓           ↓
 Raw Outputs    High-level Reco   Final Objects   Physics
 Segmentation   Particle Props     with Truth     Results
 Clustering     Energy/Direction   Matching       Plots
 Classification PID/Kinematics     Quality Flags  Tables
```

**Output Products:**
- Reconstructed particles with full kinematics
- Interaction vertices and topologies
- Energy measurements with uncertainties
- Particle identification scores
- Truth matching associations
- Quality and confidence metrics

**Key Features:**
- Modular algorithm framework
- Configurable processing chains
- Comprehensive error propagation
- Multi-threaded processing support
- Extensive validation and monitoring
- Integration with detector calibration

**Example Configuration:**
```yaml
post:
  chain:
    - name: ParticleBuilder
      track_min_length: 3.0
    - name: EnergyReconstructor
      method: 'csda'
    - name: FlashMatcher
      time_window: 10.0
```

This module is essential for converting raw ML outputs into analysis-ready
physics objects with proper uncertainties and quality assessments.
"""

from .manager import PostManager



"""Classes to process the output of dense clustering algorithms into a set
of fragments."""

import numpy as np
import torch
import torch.nn as nn

from spine.utils.cluster.dense_cluster import fit_predict, gaussian_kernel_cuda
from spine.utils.gnn.cluster import form_clusters, get_cluster_label

__all__ = ["SPICEFragmentManager"]


class SPICEFragmentManager:
    """
    Full chain model fragment mananger for SPICE Clustering
    """

    def __init__(self, frag_cfg: dict, **kwargs):
        super(SPICEFragmentManager, self).__init__(frag_cfg, **kwargs)
        self._s_thresholds = frag_cfg.get("s_thresholds", [0.0, 0.0, 0.0, 0.0])
        self._p_thresholds = frag_cfg.get("p_thresholds", [0.5, 0.5, 0.5, 0.5])
        self._spice_classes = frag_cfg.get("cluster_classes", [])
        self._spice_min_voxels = frag_cfg.get("min_voxels", 2)

    def forward(self, input, cnn_result, semantic_labels=None):
        """
        Inputs:
            - input (torch.Tensor): N x 6 (coords, edep, semantic_labels)
            - cnn_result: dict of List[torch.Tensor], containing:
                - segmentation
                - embeddings
                - seediness
                - margins

        Returns:
            - fragments
            - frag_batch_ids
            - frag_seg

        """
        if self._use_segmentation_prediction:
            assert semantic_labels is None
            semantic_labels = torch.argmax(
                cnn_result["segmentation"][0], dim=1
            ).flatten()

        batch_labels = input[:, self._batch_column]
        fragments, frag_batch_ids = [], []
        for batch_id in batch_labels.unique():
            for s in self._spice_classes:
                mask = torch.nonzero(
                    (batch_labels == batch_id) & (semantic_labels == s), as_tuple=True
                )[0]
                if len(cnn_result["embeddings"][0][mask]) < self._spice_min_voxels:
                    continue

                pred_labels = fit_predict(
                    embeddings=cnn_result["embeddings"][0][mask],
                    seediness=cnn_result["seediness"][0][mask],
                    margins=cnn_result["margins"][0][mask],
                    fitfunc=gaussian_kernel_cuda,
                    s_threshold=self._s_thresholds[s],
                    p_threshold=self._p_thresholds[s],
                )

                for c in pred_labels.unique():
                    if c < 0:
                        continue
                    fragments.append(mask[pred_labels == c])
                    frag_batch_ids.append(int(batch_id))

        fragments_np = np.empty(len(fragments), dtype=object)
        fragments_np[:] = fragments
        frag_batch_ids = np.array(frag_batch_ids)
        frag_seg = np.empty(len(fragments), dtype=np.int32)
        for i, f in enumerate(fragments):
            vals, cnts = semantic_labels[f].unique(return_counts=True)
            assert len(vals) == 1
            frag_seg[i] = vals[torch.argmax(cnts)].item()

        return fragments_np, frag_batch_ids, frag_seg



"""Functions used to manipulate a graph of nodes and edges."""

import numba as nb
import numpy as np

import spine.math as sm
from spine.data import TensorBatch
from spine.utils.globals import COORD_COLS
from spine.utils.jit import numbafy


def get_cluster_edge_features_batch(
    data, clusts, edge_index, closest_index=None, iterative=False
):
    """Batched version of :func:`get_cluster_edge_features`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    clusts : IndexBatch
        (C) List of cluster indexes
    edge_index : EdgeIndexBatch
        (2, E) Sparse incidence matrix
    closest_index : Union[np.ndarray, torch.Tensor], optional
        (C, C) : Combined index of the closest pair of voxels per edge
    iterative : bool, default False
        If `True`, uses an iterative, fast approximation for distance computations

    Returns
    -------
    TensorBatch
        (E, N_e) List of edge features between clusters
    """
    directed = edge_index.directed
    index = edge_index.index_t if directed else edge_index.directed_index_t
    counts = edge_index.counts if directed else edge_index.directed_counts
    feats = get_cluster_edge_features(
        data.tensor, clusts.index_list, index, closest_index, iterative
    )

    return TensorBatch(feats, counts)


def get_voxel_edge_features_batch(data, edge_index, max_dist=5.0):
    """Batched version of :func:`get_voxel_edge_features`.

    Parameters
    ----------
    data : TensorBatch
        Batch of cluster label data tensor
    edge_index : EdgeIndexBatch
        (2, E) Sparse incidence matrix

    Returns
    -------
    TensorBatch
        (E, N_e) List of edge features between voxels.
    """
    directed = edge_index.directed
    index = edge_index.index_t if directed else edge_index.directed_index_t
    counts = edge_index.counts if directed else edge_index.directed_counts
    feats = get_voxel_edge_features(data.tensor, index)

    return TensorBatch(feats, counts)


@numbafy(cast_args=["data"], list_args=["clusts"], keep_torch=True, ref_arg="data")
def get_cluster_edge_features(
    data, clusts, edge_index, closest_index=None, iterative=False
):
    """Returns a tensor of edge features for each edge connecting
    point clusters in the graph.

    The edge features (N_e = 19) include (in that order):
    - Coordinates of the voxel in the first cluster closest to the second (3)
    - Coordinates of the voxel in the second cluster closest to the first (3)
    - Displacement vector between the aforementioned voxels (3)
    - Magnitude of the displacement vector (1)
    - Outer product of the displacement vector (9)

    Parameters
    ----------
    data : Union[np.ndarray, torch.Tensor]
        (N, 1 + D + N_f) Batched sparse tensors
    clusts : List[np.ndarray]
        (C) List of arrays of voxels IDs in each cluster
    edge_index : Union[np.ndarray, torch.Tensor]
        (2, E) Incidence map between voxels
    closest_index : Union[np.ndarray, torch.Tensor], optional
        (C, C) : Combined index of the closest pair of voxels per edge
    iterative : bool, default False
        If `True`, uses an iterative, fast approximation for distance computations

    Returns
    -------
    np.ndarray
        (E, N_e) Tensor of edge features
    """
    if not len(clusts):
        return np.empty((0, 19), dtype=data.dtype)  # Cannot type empty list

    return _get_cluster_edge_features(
        data, clusts, edge_index, closest_index, iterative
    )
    # return _get_cluster_edge_features_vec(
    #         data, clusts, edge_index, closest_index, iterative)


@nb.njit(parallel=True, cache=True)
def _get_cluster_edge_features(
    data: nb.float32[:, :],
    clusts: nb.types.List(nb.int64[:]),
    edge_index: nb.int64[:, :],
    closest_index: nb.int64[:] = None,
    iterative: nb.boolean = False,
) -> nb.float32[:, :]:

    feats = np.empty((len(edge_index), 19), dtype=data.dtype)
    for k in nb.prange(len(edge_index)):
        # Get the voxels in the clusters connected by the edge
        c1, c2 = edge_index[k]
        x1 = data[clusts[c1]][:, COORD_COLS]
        x2 = data[clusts[c2]][:, COORD_COLS]

        # Find the closest set point in each cluster
        if closest_index is not None:
            imin = closest_index[c1, c2]
            i1, i2 = imin // len(x2), imin % len(x2)
        else:
            i1, j2, _ = sm.distance.closest_pair(x1, x2, iterative)
        v1 = x1[i1, :]
        v2 = x2[i2, :]

        # Displacement
        disp = v1 - v2

        # Distance
        lend = np.linalg.norm(disp)
        if lend > 0:
            disp = disp / lend

        # Outer product
        B = np.outer(disp, disp).flatten()

        feats[k] = np.concatenate((v1, v2, disp, np.array([lend]), B))

    return feats


@nb.njit(cache=True)
def _get_cluster_edge_features_vec(
    data: nb.float32[:, :],
    clusts: nb.types.List(nb.int64[:]),
    edge_index: nb.int64[:, :],
    closest_index: nb.int64[:] = None,
    iterative: nb.boolean = False,
) -> nb.float32[:, :]:

    # Get the closest points of approach IDs for each edge
    if closest_index is None:
        lend, idxs1, idxs2 = _get_edge_distances(
            data[:, COORD_COLS], clusts, edge_index, iterative
        )
    else:
        idxs1, idxs2 = closest_index[(edge_index[0], edge_index[1])]

    # Get the points that correspond to the first voxels
    v1 = data[idxs1][:, COORD_COLS]

    # Get the points that correspond to the second voxels
    v2 = data[idxs2][:, COORD_COLS]

    # Get the displacement
    disp = v1 - v2

    # Reshape the distance vector to a column vector
    if closest_index is None:
        lend = lend.reshape(-1, 1)
    else:
        lend = np.linalg.norm(disp, axis=1)

    # Normalize the displacement vector
    disp = disp / (lend + (lend == 0))

    # Compute the outer product of the displacement
    B = np.empty((len(disp), 9), dtype=data.dtype)
    for k in range(len(disp)):
        B[k] = np.outer(disp, disp).flatten()
    # B = np.dot(disp.reshape(len(disp),-1,1),
    #            disp.reshape(len(disp),1,-1)).reshape(len(disp),-1)

    return np.hstack((v1, v2, disp, lend, B))


@numbafy(cast_args=["data"], keep_torch=True, ref_arg="data")
def get_voxel_edge_features(data, edge_index):
    """Returns a tensor of edge features for each edge connecting
    point individual voxels in the graph.

    The edge features (N_e = 19) include (in that order):
    - Coordinates of the source voxel (3)
    - Coordinates of the target voxel (3)
    - Displacement vector between the two aforementioned voxels (3)
    - Magnitude of the displacement vector (1)
    - Outer product of the displacement vector (9)

    Parameters
    ----------
    data : Union[np.ndarray, torch.Tensor]
        (N, 1 + D + N_f) Batched sparse tensors
    clusts : List[np.ndarray]
        (C) List of arrays of voxels IDs in each cluster
    edge_index : Union[np.ndarray, torch.Tensor]
        (2, E) Incidence map between voxels

    Returns
    -------
    np.ndarray
        (E, N_e) Tensor of edge features
    """
    return _get_voxel_edge_features(data, edge_index)


@nb.njit(parallel=True, cache=True)
def _get_voxel_edge_features(
    data: nb.float32[:, :], edge_index: nb.int64[:, :]
) -> nb.float32[:, :]:
    feats = np.empty((len(edge_index), 19), dtype=data.dtype)
    for k in nb.prange(len(edge_index)):
        # Get the voxel coordinates
        xi = data[edge_index[k, 0]][:, COORD_COLS]
        xj = data[edge_index[k, 1]][:, COORD_COLS]

        # Displacement
        disp = xj - xi

        # Distance
        lend = np.linalg.norm(disp)
        if lend > 0:
            disp = disp / lend

        # Outer product
        B = np.outer(disp, disp).flatten()

        feats[k] = np.concatenate([xi, xj, disp, np.array([lend]), B])

    return feats


@numbafy(cast_args=["voxels"], list_args=["clusts"])
def get_edge_distances(voxels, clusts, edge_index, iterative):
    """For each edge, finds the closest points of approach (CPAs) between the
    two voxel clusters it connects, and the distance that separates them.

    Notes
    -----
    The voxel IDs correspond to the voxel list, not an index within a cluster.

    Parameters
    ----------
    voxels : Union[np.ndarray, torch.Tensor
        (N,3) Tensor of voxel coordinates
    clusts : List[np.ndarray]
        (C) List of arrays of voxel IDs in each cluster
    edge_index : Union[np.ndarray, torch.Tensor]
        (2, E) Incidence matrix
    iterative : bool, default False
        If `True`, uses an iterative, fast approximation for distance computations

    Returns
    -------
    np.ndarray
        (E) List of edge lengths
    np.ndarray
        (E) List of voxel IDs corresponding to the first edge cluster CPA
    np.ndarray
        (E) List of voxel IDs corresponding to the second edge cluster CPA
    """
    return _get_edge_distances(voxels, clusts, edge_index, iterative)


@nb.njit(parallel=True, cache=True)
def _get_edge_distances(
    voxels: nb.float32[:, :],
    clusts: nb.types.List(nb.int64[:]),
    edge_index: nb.int64[:, :],
    iterative: nb.boolean = False,
) -> (nb.float32[:], nb.int64[:], nb.int64[:]):

    # Loop over the provided edges
    lend = np.empty(len(edge_index), dtype=voxels.dtype)
    resi = np.empty(len(edge_index), dtype=np.int64)
    resj = np.empty(len(edge_index), dtype=np.int64)
    indxi, indxj = edge_index
    for k in nb.prange(len(indxi)):
        i, j = indxi[k], indxj[k]
        if i == j:
            ii = jj = 0
            dist = 0.0
        else:
            ii, jj, dist = sm.distance.closest_pair(
                voxels[clusts[i]], voxels[clusts[j]], iterative
            )

        lend[k] = dist
        resi[k] = clusts[i][ii]
        resj[k] = clusts[j][jj]

    return lend, resi, resj


@numbafy(cast_args=["voxels"], list_args=["clusts"])
def inter_cluster_distance(
    voxels, clusts, counts=None, centroid=False, iterative=False, return_index=False
):
    """Finds the inter-cluster distance between every pair of clusters within
    each batch, returned as a block-diagonal matrix.

    Parameters
    ----------
    voxels : Union[np.ndarray, torch.Tensor]
        (N, D) Tensor of voxel coordinates
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    counts : np.ndarray, optional
        (B) Number of clusters in each entry of the batch
    centroid : bool, default False
        If `True`, use the centroid distance as a fast, approximate proxy
    iterative : bool, default False
        If `True`, uses an iterative, fast approximation to compute voxel distance
    return_index : bool, default True
        Returns a combined index of the closest pair of voxels for each
        cluster, if the 'centroid' distance method is not used

    Returns
    -------
    Union[np.ndarray, torch.Tensor]
        (C, C) Tensor of pair-wise cluster distances
    Union[np.ndarray, torch.Tensor], optional
        (C, C) Tensor of pair-wise closest voxel pair
    """
    # If there is no counts provided, assume all clusters are in one entry
    if counts is None:
        counts = np.array([len(clusts)], dtype=np.int64)

    if not return_index:
        # If there are no clusters, return empty
        if len(clusts) == 0:
            return np.empty((0, 0), dtype=voxels.dtype)

        return _inter_cluster_distance(voxels, clusts, counts, centroid, iterative)

    else:
        # If there are no clusters, return empty
        assert not centroid, "Cannot return index for centroid method."
        if len(clusts) == 0:
            return (
                np.empty((0, 0), dtype=voxels.dtype),
                np.empty((0, 0), dtype=np.int64),
            )

        return _inter_cluster_distance_index(voxels, clusts, counts, iterative)


@nb.njit(parallel=True, cache=True)
def _inter_cluster_distance(
    voxels: nb.float32[:, :],
    clusts: nb.types.List(nb.int64[:]),
    counts: nb.int64[:],
    centroid: nb.boolean = False,
    iterative: nb.boolean = False,
) -> nb.float32[:, :]:

    # Loop over the upper diagonal elements of each block on the diagonal
    dist_mat = np.zeros((len(clusts), len(clusts)), dtype=voxels.dtype)
    indxi, indxj = complete_graph(counts)
    if not centroid:
        for k in nb.prange(len(indxi)):
            # Identifiy the two voxels closest to each other in each cluster
            i, j = indxi[k], indxj[k]
            dist_mat[i, j] = dist_mat[j, i] = sm.distance.closest_pair(
                voxels[clusts[i]], voxels[clusts[j]], iterative
            )[-1]

    else:
        # Compute the centroid of each cluster
        dtype = voxels.dtype
        centroids = np.empty((len(clusts), voxels.shape[1]), dtype=dtype)
        for i in nb.prange(len(clusts)):
            centroids[i] = sm.mean(voxels[clusts[i]], axis=0)

        # Measure the distance between cluster centroids
        for k in nb.prange(len(indxi)):
            i, j = indxi[k], indxj[k]
            dist_mat[i, j] = dist_mat[j, i] = np.sqrt(
                np.sum((centroids[j] - centroids[i]) ** 2)
            )

    return dist_mat


@nb.njit(parallel=True, cache=True)
def _inter_cluster_distance_index(
    voxels: nb.float32[:, :],
    clusts: nb.types.List(nb.int64[:]),
    counts: nb.int64[:],
    iterative: nb.boolean = False,
) -> (nb.float32[:, :], nb.int64[:, :]):

    # Loop over the upper diagonal elements of each block on the diagonal
    dist_mat = np.zeros((len(clusts), len(clusts)), dtype=voxels.dtype)
    closest_index = np.zeros((len(clusts), len(clusts)), dtype=nb.int64)
    indxi, indxj = complete_graph(counts)
    for k in nb.prange(len(indxi)):
        # Identify the two voxels closest to each other in each cluster
        i, j = indxi[k], indxj[k]
        ii, jj, dist = sm.distance.closest_pair(
            voxels[clusts[i]], voxels[clusts[j]], iterative
        )
        index = ii * len(clusts[j]) + jj

        # Store the index and the distance in a matrix
        closest_index[i, j] = closest_index[j, i] = index
        dist_mat[i, j] = dist_mat[j, i] = dist

    return dist_mat, closest_index


@numbafy(cast_args=["graph"])
def get_fragment_edges(graph, clust_ids):
    """Function that converts a set of edges between cluster ids
    to a set of edges between fragment ids (ordering in list).

    Parameters
    ----------
    graph : Union[np.ndarray, torch.Tensor]
        (E, 2) Tensor of [clust_id_1, clust_id_2]
    clust_ids : np.ndarray
        (C) List of fragment cluster ids

    Returns
    -------
    np.ndarray
        (E,2) Tensor of true edges [frag_id_1, frag_id2]
    """
    return _get_fragment_edges(graph, clust_ids)


@nb.njit(cache=True)
def _get_fragment_edges(
    graph: nb.int64[:, :], clust_ids: nb.int64[:]
) -> nb.int64[:, :]:
    # Loop over the graph edges, find the fragment ids, append
    true_edges = np.empty((0, 2), dtype=np.int64)
    for e in graph:
        n1 = np.where(clust_ids == e[0])[0]
        n2 = np.where(clust_ids == e[1])[0]
        if len(n1) and len(n2):
            true_edges = np.vstack(
                (true_edges, np.array([[n1[0], n2[0]]], dtype=np.int64))
            )

    return true_edges


@nb.njit(cache=True)
def complete_graph(counts: nb.int64[:]) -> nb.int64[:, :]:
    """Creates a list of edges corresponding to a directed complete graph
    in a batch of nodes (nodes from separate entries).

    Parameters
    ----------
    counts : np.ndarray, optional
        (B) Number of nodes in each entry of the batch
    """
    # Loop over the batches, define the adjacency matrix for each
    num_edges = np.sum((counts * (counts - 1)) // 2)
    edge_index = np.empty((2, num_edges), dtype=np.int64)
    offset, index = 0, 0
    for b in range(len(counts)):
        # Build a list of edges
        c = counts[b]
        adj_mat = np.triu(np.ones((c, c)), k=1)
        edges = np.vstack(np.where(adj_mat))
        num_edges_b = edges.shape[1]

        # Append
        edge_index[:, index : index + num_edges_b] = offset + edges
        index += num_edges_b
        offset += c

    return edge_index


@nb.njit(cache=True)
def filter_invalid_nodes(
    edge_index: nb.int64[:, :], invalid_nodes: nb.int64[:]
) -> nb.int64[:, :]:
    """Remove invalid node from a graph, bridge gaps formed by the filtering.

    Each time a node is removed, the function proceeds as follows:
    - If the node has no children, remove any edges that connect to it
    - If the node has children:
      - If it does not have a parent, remove any edges from it
      - If it has a parent, connect the parent to its children

    Parameters
    ----------
    edge_index : np.ndarray
        (E, 2) Original graph incidence map
    invalid_nodes : np.ndarray
        (N) List of nodes to remove from the original graph incidence map

    Returns
    -------
    np.ndarray
        (E', 2) Filtered graph incidence map
    """
    # Loop over the list of invalid nodes
    edges = edge_index.copy()
    for node in invalid_nodes:
        # If the node has no children, remove edges to the node
        children = np.where(edges[:, 0] == node)[0]
        if len(children) == 0:
            edges = edges[edges[:, 1] != node]
            continue

        # If it has children, find its parent and reassign its children
        parent = np.where(edges[:, 1] == node)[0]
        assert len(parent) <= 1, "Found a particle with multiple parents, not allowed."

        if len(parent) == 1:
            # If it has a parent, then assign children to that parent
            parent_id = edges[parent][0][0]
            edges[:, 0][children] = parent_id
        else:
            # If it has no parent, remove the edges to it
            edges = edges[edges[:, 0] != node]

        # Remove edges to the node
        edges = edges[edges[:, 1] != node]

    return edges



#!/usr/bin/env python3
"""Comprehensive CLI entry point that handles torch imports conditionally."""

import argparse
import os
import pathlib
import sys
from typing import List

from spine.config import load_config_file
from spine.config.loader import resolve_config_path
from spine.config.operations import parse_value, set_nested_value


def main(
    config: str,
    source: List[str],
    source_list: str,
    output: str,
    n: int,
    nskip: int,
    log_dir: str,
    weight_prefix: str,
    weight_path: str,
    config_overrides: List[str],
):
    """Main driver for training/validation/inference/analysis.

    Performs these basic functions:
    - Update the configuration with the command-line arguments
    - Run the appropriate piece of code

    Parameters
    ----------
    config : str
        Path to the configuration file
    source : List[str]
        List of paths to the input files
    source_list : str
        Path to a text file containing a list of data file paths
    output : str
        Path to the output file
    n : int
        Number of iterations to run
    nskip : int
        Number of iterations to skip
    log_dir : str
        Path to the directory for storing the training log
    weight_prefix : str
        Path to the directory for storing the training weights
    weight_path : str
        Path to a weight file or pattern for multiple weight files to load
        the model weights
    config_overrides : List[str]
        List of config overrides in the form "key.path=value"
    """
    # Load the configuration tools to find the appropriate config file
    cfg_file = resolve_config_path(config, current_dir=os.getcwd())

    # Load the configuration file using the advanced loader
    cfg = load_config_file(cfg_file)

    # If there is no base block, build one
    if "base" not in cfg:
        cfg["base"] = {}

    # Propagate the configuration parent directory to enable relative paths
    parent_path = str(pathlib.Path(cfg_file).parent)
    cfg["base"]["parent_path"] = parent_path

    # The configuration must minimally contain an IO block
    assert "io" in cfg, "Must provide an `io` block in the configuration."

    # Override the input/output command-line information into the configuration
    if source is not None and len(source) > 0:
        if "reader" in cfg["io"]:
            cfg["io"]["reader"]["file_keys"] = source
        elif "loader" in cfg["io"]:
            cfg["io"]["loader"]["dataset"]["file_keys"] = source
        else:
            raise KeyError("Must specify `loader` or `reader` in the `io` block.")

    if source_list is not None:
        if "reader" in cfg["io"]:
            cfg["io"]["reader"]["file_list"] = source_list
        elif "loader" in cfg["io"]:
            cfg["io"]["loader"]["dataset"]["file_list"] = source_list
        else:
            raise KeyError("Must specify `loader` or `reader` in the `io` block.")

    if n is not None:
        if "reader" in cfg["io"]:
            cfg["io"]["reader"]["n_entry"] = n
        elif "loader" in cfg["io"]:
            cfg["io"]["loader"]["dataset"]["n_entry"] = n
        else:
            raise KeyError("Must specify `loader` or `reader` in the `io` block.")

    if nskip is not None:
        if "reader" in cfg["io"]:
            cfg["io"]["reader"]["n_skip"] = nskip
        elif "loader" in cfg["io"]:
            cfg["io"]["loader"]["dataset"]["n_skip"] = nskip
        else:
            raise KeyError("Must specify `loader` or `reader` in the `io` block.")

    if output is not None and "writer" in cfg["io"]:
        cfg["io"]["writer"]["file_name"] = output

    if log_dir is not None:
        cfg["base"]["log_dir"] = log_dir

    if weight_prefix is not None:
        if not "train" in cfg["base"]:
            raise KeyError(
                "--weight_prefix flag provided: must specify "
                "`train` in the `base` block."
            )
        cfg["base"]["train"]["weight_prefix"] = weight_prefix

    if weight_path is not None:
        cfg["model"]["weight_path"] = weight_path

    # Apply any generic config overrides from --set arguments
    if config_overrides:
        for override in config_overrides:
            if "=" not in override:
                raise ValueError(
                    f"Invalid --set format: '{override}'. "
                    f"Expected format: 'key.path=value'"
                )

            key_path, value_str = override.split("=", 1)
            key_path = key_path.strip()
            value_str = value_str.strip()

            # Parse the value (handles strings, numbers, booleans, lists, etc.)
            value = parse_value(value_str)

            # Set the nested value (returns tuple of (config, applied))
            cfg, _ = set_nested_value(cfg, key_path, value)

    # For actual training/inference, we need the main functionality
    from spine.main import run

    # Run the main function
    run(cfg)


def cli():
    """Main CLI entry point with conditional torch imports."""
    parser = argparse.ArgumentParser(
        description="SPINE - Scalable Particle Imaging with Neural Embeddings",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  spine --version                                Show version information
  spine --info                                  Show system and dependency info
  spine -c config.cfg                           Run ML training/inference with config file
  spine -c config.cfg --set io.loader.batch_size=8    Override config parameters
  spine -c config.cfg --set base.iterations=1000 --set io.loader.batch_size=16
  spine -c config.cfg --set model.detect_anomaly=true Debug PyTorch issues
  spine --help                                  Show this help message

For ML training/inference functionality, ensure PyTorch is installed:
  pip install spine-ml[model]
""",
    )

    # Add a version command
    parser.add_argument(
        "--version", "-v", action="version", version=f"SPINE {get_version()}"
    )

    # Add basic info command
    parser.add_argument(
        "--info",
        "-i",
        action="store_true",
        help="Show system and dependency information",
    )

    # Add config file argument (-c/--config only)
    parser.add_argument(
        "-c",
        "--config",
        required=True,
        help="Path to the configuration file (requires torch dependencies)",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-s", "--source", nargs="+", type=str, help="List of paths to the input files"
    )
    group.add_argument(
        "-S",
        "--source-list",
        help="Path to a text file containing a list of data file paths",
    )

    parser.add_argument("-o", "--output", help="Path to the output file")

    parser.add_argument(
        "-n", "--iterations", type=int, help="Number of iterations to run"
    )

    parser.add_argument("--nskip", type=int, help="Number of iterations to skip")

    parser.add_argument(
        "--log-dir", help="Path to the directory for storing the training log"
    )

    parser.add_argument(
        "--weight-prefix", help="Path to the directory for storing the training weights"
    )

    parser.add_argument(
        "--weight-path",
        help="Path string a weight file or pattern for multiple weight files to load model weights",
    )

    parser.add_argument(
        "--set",
        action="append",
        dest="config_overrides",
        metavar="KEY=VALUE",
        help="Override any config parameter using dot notation "
        "(e.g., --set io.loader.batch_size=8). "
        "Can be used multiple times for multiple overrides.",
    )

    args = parser.parse_args()

    # If no arguments provided and no config, show help
    if len(sys.argv) == 1:
        parser.print_help()
        return

    # Handle version and info without torch
    if args.info:
        show_info()
        return

    config_file = args.config

    # For actual training/inference, call the main function
    main(
        config=config_file,
        source=args.source,
        source_list=args.source_list,
        output=args.output,
        n=args.iterations,
        nskip=args.nskip,
        log_dir=args.log_dir,
        weight_prefix=args.weight_prefix,
        weight_path=args.weight_path,
        config_overrides=args.config_overrides,
    )


def get_version():
    """Get SPINE version without importing heavy dependencies."""
    try:
        from spine.version import __version__

        return __version__
    except ImportError:
        return "unknown"


def show_info():
    """Show comprehensive package and system information."""
    print(f"SPINE (Scalable Particle Imaging with Neural Embeddings) v{get_version()}")
    print("https://github.com/DeepLearnPhysics/spine")
    print()

    # Check and display dependency status
    deps = check_dependencies()

    print("Dependency Status:")
    print("-" * 40)

    for name, version in deps.items():
        status = f"✓ {version}" if version else "✗ Not available"
        print(f"{name:15}: {status}")

    print(f"\nPython: {sys.version}")
    print()

    print("Available functionality:")
    print("  Core: Mathematical operations, data handling, I/O")

    if deps["torch"]:
        print(f"  Model: Neural networks available (PyTorch {deps['torch']})")
    else:
        print("  Model: Not available (install with: pip install spine-ml[model])")

    if deps["plotly"]:
        print(f"  Visualization: Available (Plotly {deps['plotly']})")
    else:
        print(
            "  Visualization: Not available (install with: pip install spine-ml[viz])"
        )

    if deps["torch"] is None:
        print("\n" + "=" * 50)
        print("NOTICE: PyTorch not found!")
        print("For full ML functionality, install with:")
        print("  pip install spine-ml[model]")
        print("=" * 50)


def check_dependencies():
    """Check what optional dependencies are available."""
    deps = {}

    # Check PyTorch
    try:
        import torch

        deps["torch"] = torch.__version__
    except ImportError:
        deps["torch"] = None

    # Check visualization dependencies
    try:
        import matplotlib

        deps["matplotlib"] = matplotlib.__version__
    except ImportError:
        deps["matplotlib"] = None

    try:
        import plotly

        deps["plotly"] = plotly.__version__
    except ImportError:
        deps["plotly"] = None

    try:
        import seaborn

        deps["seaborn"] = seaborn.__version__
    except ImportError:
        deps["seaborn"] = None

    # Check ML dependencies
    try:
        import MinkowskiEngine

        deps["minkowski"] = MinkowskiEngine.__version__
    except ImportError:
        deps["minkowski"] = None

    return deps


if __name__ == "__main__":
    cli()



"""Module which defines encoders using convolutional neural networks."""

import torch

from spine.data import IndexBatch, TensorBatch
from spine.model.layer.cnn.encoder import SparseResidualEncoder
from spine.utils.globals import BATCH_COL

__all__ = ["ClustCNNNodeEncoder", "ClustCNNEdgeEncoder", "ClustCNNGlobalEncoder"]


class ClustCNNNodeEncoder(torch.nn.Module):
    """Produces cluster node features using a sparse residual CNN encoder."""

    # Name of the node encoder (as specified in the configuration)
    name = "cnn"

    def __init__(self, **cfg):
        """Initializes the CNN-based node encoder.

        Simply passes the configuration along to the underlying sparse residual
        CNN encoder defined in :class:`SparseResidualEncoder`.

        Parameters
        ----------
        **cfg : dict, optional
            Configuration to pass along to the sparse residual encoder
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the sparse residual encoder
        self.encoder = SparseResidualEncoder(**cfg)
        self.feature_size = self.encoder.feature_size

    def forward(self, data, clusts, **kwargs):
        """Generate CNN cluster node features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            Indexes that make up each cluster
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
            (C, N_c) Set of N_c features per cluster
        """
        # Use cluster ID as a batch ID, pass through CNN
        full_index = clusts.full_index
        cnn_data = data.tensor[full_index].clone()
        cnn_data[:, BATCH_COL] = torch.tensor(clusts.index_ids, device=cnn_data.device)

        # Pass the batched input through the encoder
        feats = self.encoder(cnn_data)

        return TensorBatch(feats, clusts.counts)


class ClustCNNEdgeEncoder(torch.nn.Module):
    """Produces cluster edge features using a sparse residual CNN encoder.

    Considers an edge as an image containing both ojbects connected by
    the edge in a single image.
    """

    # Name of the edge encoder (as specified in the configuration)
    name = "cnn"

    def __init__(self, **cfg):
        """Initializes the CNN-based edge encoder.

        Simply passes the configuration along to the underlying sparse residual
        CNN encoder defined in :class:`SparseResidualEncoder`.

        Parameters
        ----------
        **cfg : dict, optional
            Configuration to pass along to the sparse residual encoder
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the sparse residual encoder
        self.encoder = SparseResidualEncoder(**cfg)
        self.feature_size = self.encoder.feature_size

    def forward(self, data, clusts, edge_index, **kwargs):
        """Generate CNN cluster edge features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            Indexes that make up each cluster
        edge_index : EdgeIndexBatch
            Incidence map between clusters
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
            (E, N_e) Set of N_e features per edge
        """
        # Use edge ID as a batch ID, pass through CNN. For undirected graph,
        # only do it on half of the edges to save time (same features).
        cnn_data = []
        for i, e in enumerate(edge_index.directed_index_t):
            ci, cj = clusts.data[e[0]], clusts.data[e[1]]
            edge_data = torch.cat((data.tensor[ci], data.tensor[cj]))
            edge_data[:, BATCH_COL] = i
            cnn_data.append(edge_data)

        # Pass through the network
        if len(cnn_data):
            feats = self.encoder(torch.cat(cnn_data))

        else:
            feats = torch.empty(
                (0, self.feature_size),
                dtype=data.tensor.dtype,
                device=data.tensor.device,
            )

        # If the graph is undirected, add reciprocal features
        if not edge_index.directed:
            full_feats = torch.empty(
                (2 * feats.shape[0], feats.shape[1]),
                dtype=feats.dtype,
                device=feats.device,
            )
            full_feats[::2] = feats
            full_feats[1::2] = feats

            feats = full_feats

        return TensorBatch(feats, edge_index.counts)


class ClustCNNGlobalEncoder(torch.nn.Module):
    """Produces graph-wide features using a sparse residual CNN encoder.

    Considers the whole graph as an image containing all objects in it.
    """

    # Name of the global encoder (as specified in the configuration)
    name = "cnn"

    def __init__(self, **cfg):
        """Initializes the CNN-based global encoder.

        Simply passes the configuration along to the underlying sparse residual
        CNN encoder defined in :class:`SparseResidualEncoder`.

        Parameters
        ----------
        **cfg : dict, optional
            Configuration to pass along to the sparse residual encoder
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the sparse residual encoder
        self.encoder = SparseResidualEncoder(**cfg)
        self.feature_size = self.encoder.feature_size

    def forward(self, data, clusts, **kwargs):
        """Generate CNN global graph features for one batch of data.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Batch of sparse tensors
        clusts : IndexBatch
            Indexes that make up each cluster
        **kwargs : dict, optional
            Additional objects no used by this encoder

        Returns
        -------
        TensorBatch
            (B, N_g) Set of N_g globale graph features per batch entry
        """
        # Restrict the set of points to those in the graph clusters
        full_index = clusts.full_index
        cnn_data = data.tensor[full_index]

        # Pass the batched input through the encoder
        feats = self.encoder(cnn_data)

        return TensorBatch(feats, [1] * clusts.batch_size)



"""Analysis script used to evaluate the clustering accuracy."""

import numpy as np

import spine.utils.metrics
from spine.ana.base import AnaBase
from spine.utils.enums import enum_factory
from spine.utils.globals import CLUST_COL, GROUP_COL, INTER_COL, LOWES_SHP, SHAPE_COL

__all__ = ["ClusterAna"]


class ClusterAna(AnaBase):
    """Class which computes and stores the necessary data to evaluate
    clustering metrics at different aggregation stages:
    - fragments
    - particles
    - interactions
    """

    # Name of the analysis script (as specified in the configuration)
    name = "cluster_eval"

    # Label column to use for each clustering label_col
    _label_cols = (
        ("fragment", CLUST_COL),
        ("particle", GROUP_COL),
        ("interaction", INTER_COL),
    )

    def __init__(
        self,
        obj_type=None,
        use_objects=False,
        per_object=True,
        per_shape=True,
        metrics=("pur", "eff", "ari"),
        label_key="clust_label_adapt",
        label_col=None,
        **kwargs,
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        obj_type : Union[str, List[str]], optional
            Name or list of names of the object types to process
        use_objects : bool, default False
            If `True`, rebuild the clustering assignments for truth and reco
            from the set of truth and reco particles
        per_object : bool, default True
            Evaluate the clustering accuracy for each object type (not relevant
            if running GrapPA standalone)
        per_shape : bool, default True
            Evaluate the clustering accuracy for each object shape (not
            relevant in the case of interactions)
        metrics : Tuple[str], default ('pur', 'eff', 'ari')
            List of clustering metrics to evaluate
        label_key : str, default 'clust_label_adapt'
            Name of the tensor which contains the cluster labels, when
            using the raw reconstruction output
        label_col : str, optional
            Column name in the label tensor specifying the aggregation label_col
        **kwargs : dict, optional
            Additional arguments to pass to :class:`AnaBase`
        """
        # Check parameters
        assert obj_type is not None or not per_object, (
            "If evaluating clustering metrics per object, provide a list "
            "of object types to evaluate the clustering for."
        )
        assert per_object or label_col is not None, (
            "If evaluating clustering standalone (not per object), must "
            "provide the name of the target clustering label column."
        )
        assert per_object or not use_objects, (
            "If evaluating clustering standalone (not per object), cannot "
            "use objects to evaluate it."
        )

        # Initialize the parent class
        super().__init__(obj_type, "both", **kwargs)

        # If the clustering is not done per object, fix target
        if not per_object:
            self.obj_type = [label_col]

        # Store the basic parameters
        self.use_objects = use_objects
        self.per_object = per_object
        self.per_shape = per_shape
        self.label_key = label_key

        # Parse the label_col column, if necessary
        self.label_col = None
        if label_col is not None:
            self.label_col = enum_factory("cluster", label_col)

        # Convert metric strings to functions
        self.metrics = {m: getattr(spine.utils.metrics, m) for m in metrics}

        # If objects are not used, remove them from the required keys
        keys = self.keys
        if not use_objects:
            for key in self.obj_keys:
                del keys[key]

        # List other necessary data products
        if self.per_object:
            if not self.use_objects:
                # Store the labels and the clusters output by the reco chain
                keys[label_key] = True
                for obj in self.obj_type:
                    keys[f"{obj}_clusts"] = True
                    if obj != "interaction":
                        keys[f"{obj}_shapes"] = True

            else:
                keys["points"] = True

        else:
            keys[label_key] = True
            keys["clusts"] = True
            keys["group_pred"] = True

        self.keys = keys

        # Initialize the output
        for obj in self.obj_type:
            self.initialize_writer(obj)

    @property
    def label_cols(self):
        """Dictionary of (key, column_id) pairs which determine which column
        in the label tensor corresponds to a specific clustering target.

        Returns
        -------
        Dict[str, int]
            Dictionary of (key, column_id) mapping from name to label column
        """
        return dict(self._label_cols)

    def process(self, data):
        """Store the clustering metrics for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over the different object types
        for obj_type in self.obj_type:
            # Build the cluster labels for this object type
            if not self.use_objects:
                # Fetch the right label column
                label_col = self.label_col or self.label_cols[obj_type]
                num_points = len(data[self.label_key])
                labels = data[self.label_key][:, label_col]
                if obj_type != "interaction":
                    shapes = data[self.label_key][:, SHAPE_COL]
                num_truth = len(np.unique(labels[labels > -1]))

            else:
                # Rebuild the labels
                num_points = len(data["points"])
                labels = -np.ones(num_points)
                num_truth = len(data[f"truth_{obj_type}s"])
                for i, obj in enumerate(data[f"truth_{obj_type}s"]):
                    labels[obj.index_adapt] = i

            # Build the cluster predictions for this object type
            preds = -np.ones(num_points)
            if self.per_object:
                shapes = -np.full(num_points, LOWES_SHP)
                if not self.use_objects:
                    # Use clusters directly from the full chain output
                    num_reco = len(data[f"{obj_type}_clusts"])
                    for i, index in enumerate(data[f"{obj_type}_clusts"]):
                        preds[index] = i
                        if obj_type != "interaction":
                            shapes[index] = data[f"{obj_type}_shapes"][i]

                else:
                    # Use clusters from the object indexes
                    num_reco = len(data[f"reco_{obj_type}s"])
                    for i, obj in enumerate(data[f"reco_{obj_type}s"]):
                        preds[obj.index] = i
                        if obj_type != "interaction":
                            shapes[obj.index] = obj.shape

            else:
                num_reco = len(data["clusts"])
                for i, index in enumerate(data["clusts"]):
                    preds[index] = data["group_pred"][i]

            # Cast the labels and predictions to int32 for metric functions
            labels = labels.astype(np.int32)
            preds = preds.astype(np.int32)

            # Evaluate clustering metrics
            row_dict = {
                "num_points": num_points,
                "num_truth": num_truth,
                "num_reco": num_reco,
            }
            for metric, func in self.metrics.items():
                valid_index = np.where((preds > -1) & (labels > -1))[0]
                row_dict[metric] = func(labels[valid_index], preds[valid_index])
                if self.per_shape and obj_type != "interaction":
                    for shape in range(LOWES_SHP):
                        shape_index = np.where((shapes == shape) & (labels > -1))[0]
                        row_dict[f"{metric}_{shape}"] = func(
                            labels[shape_index], preds[shape_index]
                        )

            self.append(obj_type, **row_dict)



"""Module with functions/classes used to identify particle species."""

import numpy as np
from scipy.optimize import minimize

from .energy_loss import (
    bethe_bloch_lar,
    bethe_bloch_mpv_lar,
    csda_ke_lar,
    csda_table_spline,
)
from .globals import KAON_PID, MUON_PID, PID_MASSES, PION_PID, PROT_PID, TRACK_SHP
from .tracking import get_track_segment_dedxs


class TemplateParticleIdentifier:
    """Class which uses dE/dx templates to do particle identification.

    The basics of template-based PID are as follows:
    - Chunk a track into chunks
    - Identify the dE/dx of the tracks in each chunk
    - Try to match the dE/dx profile to a known particle template
    """

    def __init__(
        self,
        use_table=True,
        use_mpv=False,
        include_pids=(MUON_PID, PION_PID, PROT_PID, KAON_PID),
        optimize_rr=False,
        max_rr=100,
        optimize_orient=True,
        tracking_mode="step_next",
        **tracker,
    ):
        """Initialize the template-based particle identifier.

        Parameters
        ----------
        use_table : bool, default True
            If `True`, use tabulated values of dE/dx vs residual range. The
            dE/dx is evaluated using the theoretical value otherwise.
        use_mpv : bool, default False
            If `True`, use the most-probable dE/dx value instead of the mean
        include_pids : list, default [2, 3, 4, 5]
            Particle species to consider as a viable PID candidate
        optimize_rr : bool, default False
            If `True`, vary RR to minimize the chi2 agreement. If `False`, the
            track is assumed to range out (no hard scattering, no exiting)
        max_rr : float
            Maximum allowed residual range offset
        optimize_orient : bool, default True
            If `True`, compute the chi2 w.r.t. to the start and end points of
            the track, in case the travel direction was mireconstructed
        tracking_mode : str, default 'step_next'
            Method used to chunk the track and compute the dE/dx vs RR (one of
            'step', 'step_next' or 'bin_pca')
        **tracker : dict, optional
            Arguments to pass to the tracking algorithm (track chunking)
        """
        # Store parameters
        self.use_table = use_table
        self.use_mpv = use_mpv
        self.optimize_rr = optimize_rr
        self.max_rr = max_rr
        self.optimize_orient = optimize_orient

        # If needed, initialize the spline fits to the dE/dx templates once
        self.include_pids = include_pids
        self.tables = {
            pid: csda_table_spline(pid, value="dE/dx") for pid in include_pids
        }

        # Store the tracking parameters
        # TODO: make this is a class which owns the parameters
        self.tracking_mode = tracking_mode
        self.tracking_kwargs = tracker

        # Basic logic check
        assert not use_mpv or not use_table, (
            "MPV dE/dx values are not tabulated. Must use the theoretical "
            "approach (`use_table=False`) when `use_mpv=True`."
        )

    def __call__(self, points, depositions, end_point, start_point=None):
        """Evaluate the agreement between each PID template and the particle
        track of interest.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) Track point coordinates
        depositions : np.ndarray
            (N) Deposition values in MeV
        end_point : np.ndarray
            (3) Coordinates of the end point of the track
        start_point : np.ndarray, optional
            (3) Coordinates of the start point of the track

        Returns
        -------
        int
            Best-fit particle ID
        np.ndarray
            (P) One chi2 score per particle species candidate
        """
        # Check which end of the tracks are to be considered
        end_points = [end_point]
        if self.optimize_orient:
            assert (
                start_point is not None
            ), "Must pass start_point when test both track orientations."
            end_points.append(start_point)

        # Loop over the track orientations considered
        chi2_per_pid = np.full(len(self.include_pids), -1.0)
        for epoint in end_points:
            # Segment the track, measure dE/dx and residual range values
            seg_dedxs, _, seg_rrs, _, _, _ = get_track_segment_dedxs(
                points,
                depositions,
                epoint,
                method=self.tracking_mode,
                **self.tracking_kwargs,
            )

            # Only keep segments for which the dE/dx could be measured
            mask = np.where(seg_dedxs > 0.0)[0]
            if len(mask) < 1:
                continue
            seg_dedxs, seg_rrs = seg_dedxs[mask], seg_rrs[mask]

            # Loop over PID candidates
            for i, pid in enumerate(self.include_pids):
                # Dispatch depending on the minimization method
                if not self.optimize_rr:
                    chi2 = self.chi2(seg_dedxs, seg_rrs, pid)
                else:
                    _, chi2 = self.minimize_rr(seg_dedxs, seg_rrs, pid)

                # Update chi2 for this PID
                if chi2_per_pid[i] < 0.0 or chi2 < chi2_per_pid[i]:
                    chi2_per_pid[i] = chi2

        # Return
        if np.all(chi2_per_pid == -1.0):
            return -1, chi2_per_pid
        else:
            pid = self.include_pids[np.argmin(chi2_per_pid)]
            return pid, chi2_per_pid

    def minimize_rr(self, dedxs, rrs, pid):
        """Find the residual range which minimises the chi2 fit between the
        observed dE/dx values and the spline fit to the template.

        Parameters
        ----------
        dedxs : np.ndarray
            (S) Measured values of dedxs at a set of unknown residual ranges
        rrs : np.ndarray
            (S) Residual ranges assuming the particle came to a complete stop
        pid : int
            Particle species enumerator

        Returns
        -------
        float
            Value of the residual range at the end of the track
        float
            Value of the chi2 for the optimal residual range
        """
        chi2 = lambda x: self.chi2(dedxs, x + rrs, pid)
        res = minimize(chi2, 0.0, bounds=((0.0, self.max_rr),))
        return res.x[0], res.fun

    def chi2(self, dedxs, rrs, pid):
        """Computes a chi2 score given a set of observed/expected dE/dx values.

        Parameters
        ----------
        dedxs : np.ndarray
            (S) Measured values of dedxs at a set of known residual ranges
        rrs : np.ndarray
            (S) Residual range values (one per track segment)
        pid : int
            Particle species enumerator

        Returns
        -------
        float
            Chi2 agreement value
        """
        # TODO: should get a better measure of uncertainty
        exp_dedxs = self.expected_dedxs(rrs, pid)
        diff = dedxs - exp_dedxs
        return np.sum(diff**2 / dedxs) / len(dedxs)

    def expected_dedxs(self, rrs, pid):
        """Computes the expected dE/dx values given a set of residual ranges.

        Parameters
        ----------
        rrs : np.ndarray
            (S) Residual range values (one per track segment)
        pid : int
            Particle species enumerator

        Returns
        -------
        np.ndarray
            Expected dE/dxs values from a table or the theory
        """
        if self.use_table:
            exp_dedxs = self.tables[pid](rrs)

        else:
            mass = PID_MASSES[pid]
            exp_dedxs = np.empty(len(rrs), dtype=seg_dedxs.dtype)
            for i in range(len(rrs)):
                T = csda_ke_lar(rrs[i], mass)
                if not use_mpv:
                    dedx = -bethe_bloch_lar(T, mass)
                else:
                    # TODO: get material thickness from tracker parameters
                    dedx = -bethe_bloch_mpv_lar(T, mass, 1)
                exp_dedxs[i] = dedx

        return exp_dedxs



"""Applies field non-uniformity corrections."""

__all__ = ["FieldCalibrator"]


class FieldCalibrator:
    """Applies position corrections to account for field non-uniformities
    (space charge, cathode distrotions, etc.)
    """

    name = "field"

    def __init__(self):
        """Initialize the field calibrator.

        Notes
        -----
        Placeholder until this module is implemented
        """
        raise NotImplementedError("Field calibrator not yet available.")

    def process(self):
        """Corrects for field non-uniformities.

        Notes
        -----
        Placeholder until this module is implemented
        """
        raise NotImplementedError("Field calibrator not yet available.")



"""PyTorch utilities organized by functionality.

Submodules:
- `runtime`: Tensor operations, CUDA memory, distributed operations
- `devices`: CUDA device configuration and distributed training setup
- `training`: Optimizer and learning rate scheduler factories
- `scripts`: Utility functions and PyTorch extensions (e.g., cdist_fast)
- `adabound`: Custom AdaBound optimizers (when PyTorch available)

Usage:
    from spine.utils.torch.runtime import manual_seed
    from spine.utils.torch.devices import set_visible_devices
    from spine.utils.torch.training import optim_factory
    from spine.utils.torch.scripts import cdist_fast

All utilities gracefully handle PyTorch unavailability.
"""



"""Module used to label or evaluate GNNs.

It contains two classes of functions:
- Functions used in GNN losses
- Functions used to quantify the performance of GNNs
"""

import numba as nb
import numpy as np
from scipy.sparse import csr_array
from scipy.sparse.csgraph import minimum_spanning_tree

import spine.math as sm
from spine.data import EdgeIndexBatch, IndexBatch, TensorBatch
from spine.utils.metrics import ami, ari, pur_eff, sbd

int_array = nb.int64[:]


def edge_assignment_batch(edge_index, group_ids):
    """Batched version of :func:`edge_assignment`.

    Parameters
    ----------
    edge_index: EdgeIndexBatch
        (2, E) Sparse incidence matrix
    group_ids : TensorBatch
        (C) Cluster group IDs

    Returns
    -------
    TensorBatch
        (E) Array specifying on/off edges
    """
    edge_assn = edge_assignment(edge_index.index.T, group_ids.tensor)

    return TensorBatch(edge_assn, edge_index.counts)


def edge_assignment_from_graph_batch(edge_index, true_edge_index, part_ids):
    """Batched version of :func:`edge_assignment_from_graph`.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (2, E) Input sparse incidence matrix (on clusters)
    truth_edge_index : TensorBatch
        (2, E') Label sparse incidence matrix (on particles)
    part_ids : TensorBatch
        (C) Particle IDs of the clusters

    Returns
    -------
    TensorBatch
        (E) Array specifying on/off edges
    """
    # Convert the cluster IDs in the original edge index to particle IDs
    edge_index_part = TensorBatch(
        part_ids.tensor[edge_index.index].T, edge_index.counts
    )

    # Loop over the list of entries in the batch
    edge_assn = np.zeros(edge_index.index.shape[1], dtype=np.int64)
    for b in range(clust_label.batch_size):
        lower, upper = edge_index.edges[b], edge_index.edges[b + 1]
        edge_assn[lower:upper] = edge_assignment_from_graph(
            edge_index_part[b], true_index[b]
        )

    return TensorBatch(edge_assn, edge_index.counts)


def edge_assignment_forest_batch(edge_index, edge_pred, group_ids):
    """Batched version of :func:`edge_assignment_forest`.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (2, E) Input sparse incidence matrix (on clusters)
    edge_pred : TensorBatch
        (E, 2) Logits associated with each edge
    group_ids : TensorBatch
        (C) Cluster group IDs

    Returns
    -------
    TensorBatch
        (E) Array specifying on/off edges
    TensorBatch
        (E) Array specifying edges to apply the loss to
    """
    # Loop over the list of entries in the batch
    edge_assn = np.empty(edge_index.index.shape[1], dtype=np.int64)
    valid_mask = np.empty(edge_index.index.shape[1], dtype=bool)
    for b in range(clust_label.batch_size):
        # Get the list of labels and the list of nodes to apply the loss to
        lower, upper = edge_index.edges[b], edge_index.edges[b + 1]
        edge_assn_b, edge_valid_b = edge_assignment_forest(
            edge_index[b], edge_pred[b], group_ids[b]
        )

        edge_assn[lower:upper] = edge_assn_b
        edge_valid[lower:upper] = edge_valid_b

    return (
        TensorBatch(edge_assn, counts=edge_index.counts),
        TensorBatch(edge_valid, counts=edge_valid.counts),
    )


def edge_assignment_score_batch(edge_index, edge_pred, clusts, track_node=None):
    """Batched version of :func:`edge_assignment_score`.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (2, E) Sparse incidence matrix
    edge_pred : TensorBatch
        (E, 2) Logits associated with each edge
    clusts : IndexBatch
        (C) List of cluster indexes
    track_node : TensorBatch, optional
        (C) Whether a node is a track fragment/particle or not

    Returns
    -------
    np.ndarray
        (E', 2) Optimal incidence matrix
    np.ndarray
        (C) Optimal group ID for each node
    float
        Score of the optimal incidence matrix
    """
    edge_index_list = []
    group_ids = np.empty(len(clusts.index_list), dtype=np.int64)
    scores = np.empty(edge_index.batch_size, dtype=edge_pred.dtype)
    edge_counts = np.empty(edge_index.batch_size, dtype=np.int64)
    offset = 0
    for b in range(edge_index.batch_size):
        lower, upper = clusts.edges[b], clusts.edges[b + 1]
        track_node_b = track_node[b] if track_node is not None else None
        edge_index_b, group_ids_b, score_b = edge_assignment_score(
            edge_index[b], edge_pred[b], clusts.counts[b], track_node_b
        )

        edge_index_list.append(edge_index_b + edge_index.offsets[b])
        group_ids[lower:upper] = offset + group_ids_b
        scores[b] = score_b
        edge_counts[b] = len(edge_index_b)
        if upper - lower > 0:
            offset = np.max(group_ids[lower:upper]) + 1

    # Make a new EdgeIndexBatch out of the selected edges
    new_edge_index = EdgeIndexBatch(
        np.vstack(edge_index_list).T, edge_counts, edge_index.offsets, directed=True
    )

    return new_edge_index, TensorBatch(group_ids, clusts.counts), scores


def node_assignment_batch(edge_index, edge_pred, clusts):
    """Batched version of :func:`node_assignment`.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (2, E) Sparse incidence matrix
    edge_pred : TensorBatch
        (E, 2) Logits associated with each edge
    clusts : IndexBatch
        (C) List of cluster indexes

    Returns
    -------
        np.ndarray: (C) List of group ids
    """
    # Loop over on edges, reset the group IDs of connected node
    group_ids = np.empty(len(clusts.index_list), dtype=np.int64)
    offset = 0
    for b in range(edge_index.batch_size):
        lower, upper = clusts.edges[b], clusts.edges[b + 1]
        if upper - lower > 0:
            group_ids[lower:upper] = offset + node_assignment(
                edge_index[b], edge_pred[b], clusts.counts[b]
            )
            offset = np.max(group_ids[lower:upper]) + 1

    return TensorBatch(group_ids, counts=clusts.counts)


def node_assignment_score_batch(edge_index, edge_pred, clusts, track_node=None):
    """Finds the graph that produces the lowest grouping score and use
    union-find to find group IDs for each of the nodes in the graph.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (2, E) Sparse incidence matrix
    edge_pred : TensorBatch
        (E, 2) Logits associated with each edge
    clusts : IndexBatch
        (C) List of cluster indexes
    track_node : TensorBatch, optional
        (C) Whether a node is a track fragment/particle or not

    Returns
    -------
    np.ndarray
        (C) Optimal group ID for each node
    """
    return edge_assignment_score_batch(edge_index, edge_pred, clusts, track_node)[1]


def edge_purity_mask_batch(edge_index, part_ids, group_ids, primary_ids):
    """Batched version of :func:`edge_purity_mask`.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (2, E) Sparse incidence matrix
    part_ids : TensorBatch
        (C) Array of cluster particle IDs
    group_ids : TensorBatch
        (C) Array of cluster group IDs
    primary_ids : TensorBatch
        (C) Array of cluster primary IDs

    Returns
    -------
    np.ndarray
        (E) High purity edge mask
    """
    # Loop over the entries in the batch
    valid_mask = np.empty(edge_index.index.shape[1], dtype=bool)
    for b in range(edge_index.batch_size):
        lower, upper = edge_index.edges[b], edge_index.edges[b + 1]
        valid_mask[lower:upper] = edge_purity_mask(
            edge_index[b], part_ids[b], group_ids[b], primary_ids[b]
        )

    return valid_mask


def node_purity_mask_batch(group_ids, primary_ids):
    """Batched version of :func:`node_purity_mask`.

    Parameters
    ----------
    group_ids : TensorBatch
        (C) Array of cluster group IDs
    primary_ids : TensorBatch
        (C) Cluster of cluster primary IDs

    Returns
    -------
    np.ndarray
        (C) High purity node mask
    """
    # Loop over the entries in the batch
    valid_mask = np.empty(len(group_ids.tensor), dtype=bool)
    for b in range(group_ids.batch_size):
        lower, upper = group_ids.edges[b], group_ids.edges[b + 1]
        valid_mask[lower:upper] = node_purity_mask(group_ids[b], primary_ids[b])

    return valid_mask


def primary_assignment_batch(node_pred, group_ids=None):
    """Batched version of :func:`primary_assignment`.

    Parameters
    ----------
    node_pred : TensorBatch
        (C, 2) Logits associated with each node
    group_ids : TensorBatch, optional
        (C) List of node group IDs

    Returns
    -------
    TensorBatch
        (C) Primary labels
    """
    if group_ids is None:
        primary_ids = primary_assignment(node_pred.tensor)
    else:
        primary_ids = np.empty(len(node_pred.tensor), dtype=bool)
        for b in range(node_pred.batch_size):
            lower, upper = node_pred.edges[b], node_pred.edges[b + 1]
            primary_ids[lower:upper] = primary_assignment(node_pred[b], group_ids[b])

    return TensorBatch(primary_ids, node_pred.counts)


def edge_assignment(edge_index, group_ids):
    """Determines which edges are turned on based on the group ID of the
    clusters they are connecting.

    Parameters
    ----------
    edge_index: np.ndarray
        (E, 2) Sparse incidence matrix
    group_ids : np.ndarray
        (C) Cluster group IDs

    Returns
    -------
    np.ndarray:
        (E) Array specifying on/off edges
    """
    # Set the edge as true if it connects two nodes that belong to the same
    # entry (free; no edges between entries) and the same group
    mask = group_ids[edge_index[:, 0]] == group_ids[edge_index[:, 1]]

    return mask.astype(np.int64)


def edge_assignment_from_graph(edge_index, true_edge_index, part_ids):
    """Determines which edges are turned on based on whether they appear in
    a reference list of true edges or not.

    Parameters
    ----------
    edge_index : EdgeIndexBatch
        (E, 2) Input sparse incidence matrix (on clusters)
    truth_edge_index : TensorBatch
        (E', 2) Label sparse incidence matrix (on particles)
    part_ids : np.ndarray
        (C) Particle IDs of the clusters

    Returns
    -------
    np.ndarray:
        (E) Array specifying on/off edges
    """
    # Convert the cluster IDs in the original edge index to particle IDs
    edge_index_part = part_ids[edge_index]

    # Compare with the reference sparse incidence matrix
    compare_index = lambda x, y: (x.T == y[..., None]).all(axis=1).any(axis=1)

    return compare_index(edge_index_part, true_edge_index)


def edge_assignment_forest(edge_index, edge_pred, group_ids):
    """Determines which edges must be turned on based on to form a
    minimum-spanning tree (MST) for each node group.

    For each group, find the most likely spanning tree, label the edges in the
    tree as 1. For all other edges, apply loss only if in separate groups. If
    undirected, also assign symmetric path. This method enforces that the
    network minmally forms a forest graph on the input nodes, with each tree
    in the forest spanning a target node group.

    Parameters
    ----------
    edge_index : np.ndarray
        (E, 2) Input sparse incidence matrix (on clusters)
    edge_pred : np.ndarray
        (E, 2) Logits associated with each edge
    group_ids : np.ndarray
        (C) Cluster group IDs

    Returns
    -------
    np.ndarray:
        (E) Array specifying on/off edges
    np.ndarray
        (E) Valid edge mask (edges to apply the loss to)
    """
    # If there are no edges, nothing to do here
    edge_assn = np.zeros(len(edge_index), dtype=np.int64)
    valid_mask = np.ones(len(edge_index), dtype=bool)
    if not len(edge_index):
        return edge_assn, valid_mask

    # Convert the sparse incidence matrix scores to a CSR matrix
    n = len(group_ids)
    off_scores = sm.softmax(edge_pred, axis=1)[:, 0]
    score_mat = csr_array((off_scores, edge_index.T), shape=(n, n))

    # Build the MST graph to minimize off scores
    mst_mat = minimum_spanning_tree(score_mat)
    mst_index = np.vstack(np.where(mst_mat.toarray() > 0.0))

    # Loop over the groups, turn edges on if they appear in the MST
    # TODO: understand the impact of having an undirected graph
    compare_index = lambda x, y: (x.T == y[..., None]).all(axis=1).any(axis=1)
    for g in np.unique(group_ids):
        group_index == np.where(group_ids == g)[0]
        edge_assn_g = compare_index(edge_index_b[group_index], tree_index)
        edge_assn[group_index[edge_assn_g]] = True
        edge_valid[group_index[~edge_assn_g]] = False

    return edge_assn, edge_valid


@nb.njit(cache=True)
def node_assignment(
    edge_index: nb.int64[:, :], edge_pred: nb.int64[:, :], num_nodes: nb.int64
) -> nb.int64[:]:
    """Assigns each node to a group, based on the edge assigment provided.

    This uses the locally-defined union find implementation.

    Parameters
    ----------
    edge_index : np.ndarray
        (2, E) Sparse incidence matrix
    edge_pred : np.ndarray
        (E, 2) Logits associated with each edge
    num_nodes : int
        Number of nodes in the graph, C

    Returns
    -------
    np.ndarray
        (C) Assigned node group IDs
    """
    # Loop over on edges, reset the group IDs of connected node
    on_edges = edge_index[np.where(edge_pred[:, 1] > edge_pred[:, 0])[0]]

    return sm.graph.union_find(on_edges, num_nodes, return_inverse=True)[0]


@nb.njit(cache=True)
def node_assignment_bipartite(
    edge_index: nb.int64[:, :],
    edge_label: nb.int64[:],
    primaries: nb.int64[:],
    num_nodes: nb.int64,
) -> nb.int64[:]:
    """Assigns each node to a group represented by a primary node.

    This function loops over secondaries and associates it to the primary with
    that is connected to it with the strongest edge.

    Parameters
    ----------
    edge_index : np.ndarray
        (2, E) Sparse incidence matrix
    edge_pred : np.ndarray
        (E, 2) Logits associated with each edge
    primaries : np.ndarray
        (P) List of primary IDs
    num_nodes : int
        Number of nodes in the graph, C

    Returns
    -------
    np.ndarray
        (C) Assigned node group IDs
    """
    group_ids = np.arange(num_nodes, dtype=np.int64)
    others = [i for i in range(num_nodes) if i not in primaries]
    for i in others:
        inds = edge_index[:, 1] == i
        if np.sum(inds) == 0:
            continue
        indmax = np.argmax(edge_label[inds])
        group_ids[i] = edge_index[inds, 0][indmax]

    return group_ids


@nb.njit(cache=True)
def primary_assignment(
    node_pred: nb.float32[:, :], group_ids: nb.int64[:] = None
) -> nb.boolean[:]:
    """Select shower primary fragments based on the node-score.

    If node groupings are provided, selects a single primary per node
    group: the one that is most likely.

    Parameters
    ----------
    node_pred : np.ndarray
        (C, 2) Logits associated with each node
    group_ids : np.ndarray, optional
        (C) List of node group IDs

    Returns
    -------
    np.ndarray
        (C) Primary labels
    """
    if group_ids is None:
        return sm.argmax(node_pred, axis=1).astype(np.bool_)

    primary_ids = np.zeros(len(node_pred), dtype=np.bool_)
    node_pred = sm.softmax(node_pred, axis=1)
    for g in np.unique(group_ids):
        mask = np.where(group_ids == g)[0]
        idx = np.argmax(node_pred[mask][:, 1])
        primary_ids[mask[idx]] = True

    return primary_ids


@nb.njit(cache=True)
def adjacency_matrix(edge_index: nb.int64[:, :], n: nb.int64) -> nb.boolean[:, :]:
    """Creates a dense adjacency matrix from a list of connected edges in a
    graph, i.e. densify the graph incidence matrix.

    Parameters
    ----------
    edge_index : np.ndarray
        (2, E) Sparse incidence matrix
    num_nodes : int
        Number of nodes in the graph, C

    Returns
    -------
    np.ndarray
        (C, C) Adjacency matrix
    """
    # Cannot use double array indexing to fill the matrix
    adj_mat = np.eye(n, dtype=np.bool_)
    for i, j in edge_index:
        adj_mat[i, j] = True

    return adj_mat


@nb.njit(cache=True)
def grouping_loss(
    pred_mat: nb.float32[:], target_mat: nb.boolean[:], loss: str = "ce"
) -> np.float32:
    """Defines the graph clustering score.

    Given a target adjacency matrix A and a predicted adjacency P, the score is
    evaluated the average CE, L1 or L2 distance between truth and prediction.

    Parameters
    ----------
    pred_mat : np.ndarray
        (C*C) Predicted adjacency matrix scores (flattened)
    target_mat : np.ndarray
        (C*C) Target adjacency matrix scores (flattened)
    loss : str, default 'ce'
        Loss mode used to compute the graph score

    Returns
    -------
    float
        Graph grouping loss
    """
    if loss == "ce":
        return sm.log_loss(target_mat, pred_mat)
    elif loss == "l1":
        return np.mean(np.absolute(pred_mat - target_mat))
    elif loss == "l2":
        return np.mean((pred_mat - target_mat) * (pred_mat - target_mat))
    else:
        raise ValueError("Loss type not recognized")


@nb.njit(cache=True)
def edge_assignment_score(
    edge_index: nb.int64[:, :],
    edge_pred: nb.float32[:, :],
    num_nodes: nb.int64,
    track_node: nb.boolean[:] = None,
) -> (nb.int64[:, :], nb.int64[:], nb.float32):
    """Finds the graph that produces the lowest grouping score by iteratively
    adding the next most likely edge, if it improves the the score. This method
    effectively builds a spanning tree.

    Parameters
    ----------
    edge_index : np.ndarray
        (2, E) Sparse incidence matrix
    edge_pred : np.ndarray
        (E, 2) Logits associated with each edge
    num_nodes : int
        Number of nodes in the graph, C
    track_node : np.ndarray, optional
        (C) Whether a node is a track fragment/particle or not

    Returns
    -------
    np.ndarray
        (E', 2) Optimal incidence matrix
    np.ndarray
        (C) Optimal group ID for each node
    float
        Score of the optimal incidence matrix
    """
    # If there is no edge, do not bother
    if not len(edge_index):
        return (
            np.empty((0, 2), dtype=np.int64),
            np.arange(num_nodes, dtype=np.int64),
            0.0,
        )

    # Build an input adjacency matrix to constrain the edge selection to
    # the input graph
    adj_mat = adjacency_matrix(edge_index, num_nodes)

    # Interpret the softmax score as a dense adjacency matrix probability
    edge_scores = sm.softmax(edge_pred, axis=1)
    pred_adj = np.eye(num_nodes, dtype=edge_pred.dtype)
    for k, (i, j) in enumerate(edge_index):
        pred_adj[i, j] = edge_scores[k, 1]

    # Remove edges with a score < 0.5 and sort the remainder by increasing
    # order of OFF score
    on_mask = edge_scores[:, 1] >= 0.5
    args = np.argsort(edge_scores[on_mask, 0])
    ord_index = edge_index[on_mask][args]

    # Now iteratively identify the best edges, until the total score cannot
    # be improved any longer
    empty_adj = np.eye(num_nodes, dtype=np.bool_)
    best_ids = np.empty(0, dtype=np.int64)
    best_groups = np.arange(num_nodes, dtype=np.int64)
    track_used = np.zeros(num_nodes, dtype=np.bool_)
    best_loss = grouping_loss(pred_adj.flatten(), empty_adj.flatten())
    known_pairs = nb.typed.List.empty_list(nb.int64)
    for k, (a, b) in enumerate(ord_index):
        # If the edge connect two nodes already in the same group, proceed
        group_a, group_b = best_groups[a], best_groups[b]
        if group_a == group_b:
            continue

        # If the group pair has already been checked against, proceed
        pair = (min(group_a, group_b), max(group_a, group_b))
        pair_hash = pair[0] * num_nodes + pair[1]
        if pair_hash in known_pairs:
            continue

        # If requested, check whether there is already a track connection
        if track_node is not None:
            if track_used[a] or track_used[b]:
                continue
            if track_node[a] ^ track_node[b]:
                if not track_node[a]:
                    track_used[a] = True
                if not track_node[b]:
                    track_used[b] = True

        # Restrict the adjacency matrix and the predictions to the nodes in
        # the two candidate groups
        node_mask = np.where((best_groups == group_a) | (best_groups == group_b))[0]
        sub_pred = sm.linalg.submatrix(pred_adj, node_mask, node_mask).flatten()
        sub_adj = sm.linalg.submatrix(adj_mat, node_mask, node_mask).flatten()

        # Compute the current adjacency matrix between the two groups
        current_adj = (
            best_groups[node_mask] == best_groups[node_mask].reshape(-1, 1)
        ).flatten()

        # Join the two groups if it minimizes the loss
        current_loss = grouping_loss(sub_pred, sub_adj * current_adj)
        combined_loss = grouping_loss(sub_pred, sub_adj)
        if combined_loss < current_loss:
            best_groups[best_groups == pair[1]] = pair[0]
            best_loss += combined_loss - current_loss
            best_ids = np.append(best_ids, k)
            for pair_hash in list(known_pairs):
                pair_i = (pair_hash // num_nodes, pair_hash % num_nodes)
                if group_a in pair_i or group_b in pair_i:
                    known_pairs.remove(pair_hash)
        else:
            known_pairs.append(pair_hash)

    # Build the edge index
    best_index = ord_index[best_ids]

    return best_index, best_groups, best_loss


@nb.njit(cache=True)
def node_assignment_score(
    edge_index: nb.int64[:, :],
    edge_pred: nb.float32[:, :],
    num_nodes: nb.int64,
    track_node: nb.boolean[:] = None,
) -> nb.int64[:]:
    """Finds the graph that produces the lowest grouping score and use
    union-find to find group IDs for each of the nodes in the graph.

    Parameters
    ----------
    edge_index : np.ndarray
        (2, E) Sparse incidence matrix
    edge_pred : TensorBatch
        (E, 2) Logits associated with each edge
    num_nodes : int
        Number of nodes in the graph, C
    track_node : np.ndarray, optional
        (C) Whether a node is a track fragment/particle or not

    Returns
    -------
    np.ndarray
        (C) Optimal group ID for each node
    """
    return edge_assignment_score(edge_index, edge_pred, num_nodes, track_node)[1]


@nb.njit(cache=True)
def node_purity_mask(group_ids: nb.int64[:], primary_ids: nb.int64[:]) -> nb.boolean[:]:
    """Creates a mask that is `True` only for node which belong to a group
    with more exactly one primary.

    This is useful for shower clustering only, for which there can be no or
    multiple primaries in the group, making the primary identification
    ill-defined.

    Note: It is possible that the single true primary has been broken into
    several nodes. In that case, the primary is also ambiguous, skip.
    TODO: pick the most sensible primary in that case, too restrictive
    otherwise (complicated, though).

    Parameters
    ----------
    group_ids : np.ndarray
        (C) Array of cluster group IDs
    primary_ids : np.ndarray
        (C) Cluster of cluster primary IDs

    Returns
    -------
    np.ndarray
        (C) High purity node mask
    """
    purity_mask = np.ones(len(group_ids), dtype=np.bool_)
    for g in np.unique(group_ids):
        group_mask = np.where(group_ids == g)[0]
        if np.sum(primary_ids[group_mask] == 1) != 1:
            purity_mask[group_mask] = False

    return purity_mask


@nb.njit(cache=True)
def edge_purity_mask(
    edge_index: nb.int64[:, :],
    part_ids: nb.int64[:],
    group_ids: nb.int64[:],
    primary_ids: nb.int64[:],
) -> nb.boolean[:]:
    """Creates a mask that is `True` only for edges which connect two nodes
    that both belong to a common group which has a single clear primary.

    This is useful for shower clustering only, for which there can be no or
    multiple primaries in the group, making the the edge classification
    ill-defined (no primary typically indicates a shower which originates
    outside of the active volume).

    Note: It is possible that the single true primary has been broken into
    several nodes; that is acceptable.

    Parameters
    ----------
    edge_index : np.ndarray
        (2, E) Sparse incidence matrix
    part_ids : np.ndarray
        (C) Array of cluster particle IDs
    group_ids : np.ndarray
        (C) Array of cluster group IDs
    primary_ids : np.ndarray
        (C) Array of cluster primary IDs

    Returns
    -------
    np.ndarray
        (E) High purity edge mask
    """
    # Start by building a mask of valid nodes
    node_purity_mask = np.ones(len(group_ids), dtype=np.bool_)
    for g in np.unique(group_ids):
        group_mask = np.where(group_ids == g)[0]
        primary_ids_g = primary_ids[group_mask]
        part_ids_g = part_ids[group_mask]
        if len(np.unique(part_ids_g[primary_ids_g == 1])) != 1:
            # If there not exactly one primary particle ID, the group
            # is not valid
            node_purity_mask[group_mask] = False

    # Edges that connect two invalid nodes are invalid
    purity_mask = (
        node_purity_mask[edge_index[:, 0]] | node_purity_mask[edge_index[:, 1]]
    )

    return purity_mask


def clustering_metrics(clusts, node_assn, node_pred):
    """Computes several clustering metrics for a set of clusters.

    Parameters
    ----------
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    node_assn : np.ndarray
        (C) True node groups labels
    node_pred : np.ndarray
        (C) Predicted node group labels

    Returns
    -------
    float
        Adjusted Rand Index (ARI)
    float
        Adjusted Mutual information (AMI)
    float
        Symetric Best Dice (SBD)
    float
        Purity
    float
        Efficiency
    """
    pred_vox = cluster_to_voxel_label(clusts, node_pred)
    true_vox = cluster_to_voxel_label(clusts, node_assn)
    ari_val = ari(truth_vox, pred_vox)
    ami_val = ami(truth_vox, pred_vox)
    sbd_val = sbd(truth_vox, pred_vox)
    pur_val, eff_val = pur_eff(truth_vox, pred_vox)

    return ari_val, ami_val, sbd_val, pur_val, eff_val


def voxel_efficiency_bipartite(clusts, node_assn, node_pred, primaries):
    """Computes the fraction of secondary voxels that are associated to the
    correct primary.

    Parameters
    ----------
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    node_assn : np.ndarray
        (C) True node groups labels
    node_pred : np.ndarray
        (C) Predicted node group labels
    node_pred : np.ndarray
        (P) List of primary IDs

    Returns
    -------
    float
        Fraction of correctly assigned secondary voxels
    """
    others = [i for i in range(n) if i not in primaries]
    tot_vox = np.sum([len(clusts[i]) for i in others])
    int_vox = np.sum([len(clusts[i]) for i in others if node_pred[i] == node_assn[i]])

    return int_vox / tot_vox


@nb.njit(cache=True)
def cluster_to_voxel_label(
    clusts: nb.types.List(nb.int64[:]), node_labels: nb.int64[:]
) -> nb.int64[:]:
    """Turns a list of labels on clusters to an array of labels on voxels.

    Parameters
    ----------
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    node_labels : np.ndarray
        (C) Node labels

    Returns
    -------
    np.ndarray
        (N) Voxel labels
    """
    counts = [len(c) for c in clusts]

    return np.repeat(node_labels, counts)



import MinkowskiEngine as ME
import torch
import torch.nn as nn

from spine.model.layer.cnn.act_norm import act_factory, norm_factory
from spine.model.layer.cnn.blocks import DropoutBlock, ResNetBlock
from spine.model.layer.cnn.configuration import setup_cnn_configuration


class MCDropoutDecoder(torch.nn.Module):
    """
    Convolutional decoder with dropout layers.

    The architecture is exactly the same as the ME ResidualEncoders,
    except for the additional DropoutBlocks

    Attributes:

        dropout_p: dropping probability value for dropout layers
        dropout_layer_index: layer numbers to swap resnet blocks with
        dropout resnet blocks.
    """

    def __init__(self, cfg, name="mcdropout_decoder"):
        super(MCDropoutDecoder, self).__init__()
        setup_cnn_configuration(self, cfg, "network_base")

        # UResNet Configurations
        self.model_config = cfg.get(name, {})
        self.reps = self.model_config.get("reps", 2)
        self.kernel_size = self.model_config.get("kernel_size", 2)
        self.depth = self.model_config.get("depth", 5)
        self.num_filters = self.model_config.get("num_filters", 16)
        self.nPlanes = [i * self.num_filters for i in range(1, self.depth + 1)]
        self.downsample = [self.kernel_size, 2]

        self.enc_nfilters = self.model_config.get("encoder_num_filters", None)
        if self.enc_nfilters is None:
            self.enc_nfilters = self.num_filters
        self.encoder_nPlanes = [i * self.enc_nfilters for i in range(1, self.depth + 1)]
        self.nPlanes[-1] = self.encoder_nPlanes[-1]

        self.dropout_p = self.model_config.get("dropout_p", 0.5)
        self.dropout_layer_index = self.model_config.get(
            "dropout_layers", set([i for i in range(self.depth // 2, self.depth)])
        )

        # Initialize Decoder
        self.decoding_block = []
        self.decoding_conv = []
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(norm_factory(self.norm, self.nPlanes[i + 1], **self.norm_args))
            m.append(act_factory(self.activation_name, **self.activation_args))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.nPlanes[i + 1],
                    out_channels=self.nPlanes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.D,
                )
            )
            if i in self.dropout_layer_index:
                m.append(ME.MinkowskiDropout(p=self.dropout_p))
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                if i in self.dropout_layer_index:
                    m.append(
                        DropoutBlock(
                            self.nPlanes[i] * (2 if j == 0 else 1),
                            self.nPlanes[i],
                            dimension=self.D,
                            p=self.dropout_p,
                            activation=self.activation_name,
                            activation_args=self.activation_args,
                            normalization=self.norm,
                            normalization_args=self.norm_args,
                            bias=self.allow_bias,
                        )
                    )
                else:
                    m.append(
                        ResNetBlock(
                            self.nPlanes[i] * (2 if j == 0 else 1),
                            self.nPlanes[i],
                            dimension=self.D,
                            activation=self.activation_name,
                            activation_args=self.activation_args,
                            normalization=self.norm,
                            normalization_args=self.norm_args,
                            bias=self.allow_bias,
                        )
                    )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)

    def forward(self, final, encoderTensors):
        """
        Vanilla UResNet Decoder
        INPUTS:
            - encoderTensors (list of SparseTensor): output of encoder.
        RETURNS:
            - decoderTensors (list of SparseTensor):
            list of feature tensors in decoding path at each spatial resolution.
        """
        decoderTensors = []
        x = final
        for i, layer in enumerate(self.decoding_conv):
            eTensor = encoderTensors[-i - 2]
            x = layer(x)
            x = ME.cat(eTensor, x)
            x = self.decoding_block[i](x)
            decoderTensors.append(x)
        return decoderTensors
